# plugs/wowwiki.py
#
#

__copyright__ = 'this file is in the public domain'

from gozerbot.generic import geturl, striphtml, splittxt, handle_exception, \
fromenc
from gozerbot.commands import cmnds
from gozerbot.aliases import aliases
from gozerbot.examples import examples
from gozerbot.utils.rsslist import rsslist
from gozerbot.plughelp import plughelp
from urllib import quote
import re

plughelp.add('wowwiki', 'query wowwiki')

wikire = re.compile('start content(.*?)end content', re.M)

def getwikidata(url, ievent):
    """ fetch wiki data """
    try:
        result = fromenc(geturl(url))
    except IOError, ex:
        try:
            errno = ex[0]
        except IndexError:
            handle_exception(ievent=ievent)
            return
        ievent.reply('invalid option')
        return
    if not result:
        ievent.reply("can't find data for %s" % url)
        return
    res = rsslist(result)
    txt = ""
    for i in res:
        try:
            txt = i['text']
            break
        except:
            pass
    if not txt:
        ievent.reply("no data found on %s" % url)
        return
    #txt = re.sub('\[\[Image:([^\[\]]+|\[\[[^\]]+\]\])*\]\]', '', txt)
    txt = txt.replace('[[', '')
    txt = txt.replace(']]', '')
    txt = re.sub('\s+', ' ', txt)
    return txt

def handle_wowwiki(bot, ievent):
    """ wikipedia <what> .. search wikipedia for <what> """
    if not ievent.rest:
        ievent.missing('<what>')
        return
    what = ""
    lang = 'en'
    for i in ievent.rest.split():
        first = i[0].upper()
        rest = i[1:]
        if i.startswith('-'):
            if len(i) != 3:
                ievent.reply('invalid option')
                return
            lang = i[1:]
            continue
        what += "%s%s " % (first, rest)
    what = what.strip().replace(' ', '_')
    url = 'http://wowwiki.com/wiki/Special:Export/%s' % quote(what.encode('utf-8'))
    url2 = 'http://wowwiki.com/wiki/%s' % quote(what.encode('utf-8'))
    txt = getwikidata(url, ievent)
    if not txt:
        return
    if '#REDIRECT' in txt or '#redirect' in txt:
        redir = ' '.join(txt.split()[1:])
        url = 'http://wowwiki.com/wiki/Special:Export/%s' % quote(redir.encode('utf-8'))
        url2 = 'http://wowwiki.com/wiki/%s' % quote(redir.encode('utf-8'))
        txt = getwikidata(url, ievent)
    if not txt:
        return
    res = ['%s ===> ' % url2, ]
    res += splittxt(striphtml(txt).strip())
    ievent.reply(res)

cmnds.add('wowwiki', handle_wowwiki, 'USER')
examples.add('wowwiki', 'wowwiki <what> .. search wowwiki for <what>','1) wowwiki \
gozerbot')
aliases.data['wow'] = 'wowwiki'



""" show weather based on Google's weather API """

__copyright__ = 'this file is in the public domain'
__author__ = 'Landon Fowles'

from gozerbot.generic import geturl, waitforuser, getwho, rlog
from gozerbot.commands import cmnds
from gozerbot.datadir import datadir
from gozerbot.persist.persist import Persist
from gozerbot.users import users
from gozerbot.plughelp import plughelp
from gozerbot.examples import examples
from gozerbot.persist.persiststate import UserState
from gozerbot.aliases import aliasset
import time
from xml.dom import minidom
from urllib import urlencode

plughelp.add('weather', 'show weather by zipcode/city name')

def handle_weather(bot, ievent):
    """ show weather using Google's weather API """
    userhost = ""
    loc = ""
    try:
        nick = ievent.options['--u']
        if nick:
            userhost = getwho(bot, nick)
            if not userhost:
                ievent.reply("can't determine username for %s" % nick)
                return
            else:
                try:
                    name = users.getname(userhost)
                    if not name:
                         ievent.reply("%s is not known with the bot" % nick)
                         return
                    us = UserState(name)
                    loc = us['location']
                except KeyError:
                    ievent.reply("%s doesn't have his location set in \
userstate" % nick)
                    return
    except KeyError:
         pass
    if not loc:
        if ievent.rest:
             loc = ievent.rest
        else:
             ievent.missing('--u <nick> or <location>')
             return
    query = urlencode({'weather':loc})
    weathertxt = geturl('http://www.google.ca/ig/api?%s' % query)
    if 'problem_cause' in weathertxt:
        rlog(10, 'weather', 'ERROR: %s' % weathertxt)
        ievent.reply('an error occured looking up data for %s' % loc)
        return
    resultstr = ""
    if len(weathertxt) > 135:
        gweather = minidom.parseString(weathertxt)
        gweather = gweather.getElementsByTagName('weather')[0]
        if ievent.command == "weather":
            info = gweather.getElementsByTagName('forecast_information')[0]
            if info:
                city = info.getElementsByTagName('city')[0].attributes["data"].value
                zip = info.getElementsByTagName('postal_code')[0].attributes["data"].value
                time = info.getElementsByTagName('current_date_time')[0].attributes["data"].value

                weather = gweather.getElementsByTagName('current_conditions')[0]
                condition = weather.getElementsByTagName('condition')[0].attributes["data"].value
                temp_f = weather.getElementsByTagName('temp_f')[0].attributes["data"].value
                temp_c = weather.getElementsByTagName('temp_c')[0].attributes["data"].value
                humidity = weather.getElementsByTagName('humidity')[0].attributes["data"].value
                wind = weather.getElementsByTagName('wind_condition')[0].attributes["data"].value

                try:
                    wind_km = round(int(wind[-6:-4]) * 1.609344)
                except ValueError:
                    wind_km = ""

                if (not condition == ""):
                    condition = " Oh, and it's " + condition + "."

                resultstr = "As of %s, %s (%s) has a temperature of %sC/%sF with %s. %s (%s km/h).%s" % (time, city, zip, temp_c, temp_f, humidity, wind, wind_km, condition)
        elif ievent.command == "forecast":
            forecasts = gweather.getElementsByTagName('forecast_conditions')
            for forecast in forecasts:
                condition = forecast.getElementsByTagName('condition')[0].attributes["data"].value
                low_f = forecast.getElementsByTagName('low')[0].attributes["data"].value
                high_f = forecast.getElementsByTagName('high')[0].attributes["data"].value
                day = forecast.getElementsByTagName('day_of_week')[0].attributes["data"].value
                low_c = round((int(low_f) - 32) * 5.0 / 9.0)
                high_c = round((int(high_f) - 32) * 5.0 / 9.0)
                resultstr += "[%s: F(%sl/%sh) C(%sl/%sh) %s]" % (day, low_f, high_f, low_c, high_c, condition)
    if not resultstr:
        ievent.reply('%s not found!' % loc)
        return
    else:
        ievent.reply(resultstr)

cmnds.add('weather', handle_weather, 'USER', options={'--u': ''})
aliasset('forecast', 'weather')
examples.add('weather', 'get weather for <LOCATION> or <nick>', '1) weather London, \
England 2) weather dunker')



# plugins/autoreply.py
#
#

""" do autoreply on incoming jabber private messages except commands. to enable
 use autoreply-cfg txt <message to send>. """

__copyright__ = 'this file is in the public domain'

from gozerbot.commands import cmnds
from gozerbot.callbacks import jcallbacks
from gozerbot.aliases import aliasset
from gozerbot.persist.persistconfig import PersistConfig
from gozerbot.plughelp import plughelp
from gozerbot.examples import examples
from gozerbot.tests import tests

cfg = PersistConfig()
cfg.define('txt', "")

plughelp.add('autoreply', 'jabber autoreply')

def preautoreply(bot, ievent):
    """ check where autoreply callbacks should fire """
    if not ievent.usercmnd and cfg.get('txt') and not ievent.groupchat:
        return 1

def cbautoreply(bot, ievent):
    """ do the auto reply """
    bot.say(ievent.userhost, cfg.get('txt'))

jcallbacks.add('Message', cbautoreply, preautoreply)

def handle_autoreplydisable(bot, ievent):
    """ disable autoreply """
    cfg.set('txt', '')
    ievent.reply('autoreply is disabled')
    
cmnds.add('autoreply-disable', handle_autoreplydisable, 'OPER')
examples.add('autoreply-disable', 'disable the autoreply functionality', \
'autoreply-disable')

aliasset('autoreply-set', 'autoreply-cfg txt')
aliasset('autoreply-enable', 'autoreply-cfg txt')
tests.add('autoreply-enable mekker', 'set').add('autoreply-disable')



# -*- coding: utf-8 -*-
# Support for bugtrackers
# Wijnand 'tehmaze' Modderman - http://tehmaze.com
# BSD License
#
# ======
#  TODO
# ======
#  * add authorization support to bugtrackers
#  * add write support to bugtrackers (close/open bug)
#
# ======
#  BUGS 
# ======
#  * I did test these bugtrackers on various installations, although not everything
#    will work on all installations, for example on older versions or on bugtrackers
#    that restyled the complete layout

from gozerbot.aliases import aliases
from gozerbot.callbacks import callbacks
from gozerbot.commands import cmnds
from gozerbot.datadir import datadir
from gozerbot.examples import examples
from gozerbot.generic import geturl, geturl2, striphtml, useragent
from gozerbot.persist.pdod import Pdod
from gozerbot.persist.persistconfig import PersistConfig
from gozerbot.plughelp import plughelp
from gozerbot.users import users
from gozerbot.utils.generic import convertpickle
import copy
import os
import re
import urllib, urllib2
import xml.dom.minidom

plughelp.add('bugtracker', 'various bug tracker support, define a bugtracker for the channel with bugtracker-set')

cfg = PersistConfig()
cfg.define('sep', '..')

class BugTrackerNotFound(Exception):
    pass

class BugTrackerNotImplemented(Exception):
    pass

class BugTrackerNotSupported(Exception):
    pass

class BugTracker:

    bugurls = [] # regexp patterns that could match a bugtracker url
    bugtags = re.compile('(?:bug|ticket)\s?#(\d+)', re.I)

    def __init__(self, baseUrl, autoshow):
        self.url = baseUrl.rstrip('/')
        self.autoshow = autoshow

    def close(self, bugId, userhost='', message='', action=''):
        raise BugTrackerNotImplemented

    def close_url(self, bugId):
        raise BugTrackerNotImplemented

    def comments(self, bugId):
        raise BugTrackerNotImplemented

    def comments_url(self, bugId):
        raise BugTrackerNotImplemented

    def list(self):
        raise BugTrackerNotImplemented

    def list_url(self):
        raise BugTrackerNotImplemented

    def open(self, message=''):
        raise BugTrackerNotImplemented

    def open_url(self):
        raise BugTrackerNotImplemented

    def show(self, bugId):
        raise BugTrackerNotImplemented

    def show_url(self, bugId):
        raise BugTrackerNotImplemented
        
    def handle_close(self, bot, ievent):
        if len(ievent.args) < 3:
            ievent.missing('<bug id> <action> <message>')
            return
        try:
            status = self.close(ievent.args[0], users.getname(ievent.userhost), 
                ' '.join(ievent.args[2:]), ievent.args[1])
            ievent.reply('ok')
        except AssertionError, e:
            ievent.reply('error: %s' % e)
        except BugTrackerNotImplemented:
            ievent.reply('error: not implemented in this bug tracker')
        except BugTrackerNotSupported, e:
            ievent.reply('error: not supported: %s' % e)
        except BugTrackerNotFound, e:
            ievent.reply('error: not found: %s' % e)

    def handle_comments(self, bot, ievent):
        if not ievent.args or not ievent.args[0].isdigit():
            ievent.missing('<bug id>')
            return
        try:
            comments = self.comments(ievent.args[0])
            if comments:
                ievent.reply('comments: ', comments, dot=' %s ' % cfg.get('sep'), nr=True, nritems=True)
            else:
                ievent.reply('no comments found')
        except AssertionError, e:
            ievent.reply('error: %s' % e)
        except BugTrackerNotImplemented:
            ievent.reply('error: not implemented in this bug tracker')
        except BugTrackerNotSupported, e:
            ievent.reply('error: not supported: %s' % e)
        except BugTrackerNotFound, e:
            ievent.reply('error: not found: %s' % e)

    def handle_list(self, bot, ievent):
        try:
            bugs = self.list()
            if bugs:
                ievent.reply('open bugs: ', bugs, dot=' %s ' % cfg.get('sep'))
            else:
                ievent.reply('no open bugs found!')
        except BugTrackerNotImplemented:
            ievent.reply('error: not implemented in this bug tracker')
        except BugTrackerNotSupported, e:
            ievent.reply('error: not supported: %s' % e)
        except BugTrackerNotFound, e:
            ievent.reply('error: not found: %s' % e)

    def handle_show(self, bot, ievent):
        if len(ievent.args) != 1:
            try:
                open_url = self.open_url()
                ievent.missing('<bug id> -- if you were looking to report a bug, please go to %s' % open_url)
            except BugTrackerNotImplemented:
                ievent.missing('<bug id>')
            return
        try:
            status = self.show(ievent.args[0])
            result = []
            keys   = status.keys() ; keys.sort()
            for key in keys:
                if status[key]:
                    result.append('%s=%s' % (key, status[key]))
            result.append(self.show_url(ievent.args[0]))
            ievent.reply(result, dot=' %s ' % cfg.get('sep'))
        except AssertionError, e:
            ievent.reply('error: %s' % e)
        except BugTrackerNotImplemented:
            ievent.reply('error: not implemented in this bug tracker')
        except BugTrackerNotSupported, e:
            ievent.reply('error: not supported: %s' % e)
        except BugTrackerNotFound, e:
            ievent.reply('error: not found: %s' % e)

    def scan_privmsg(self, bot, ievent):
        if not self.autoshow:
            return
        if not self.bugurls:
            return
        for bugurl in self.bugurls:
            test = bugurl.findall(ievent.origtxt)
            if test:
                nevent = copy.copy(ievent)
                nevent.args = [test[0]]
                self.handle_show(bot, nevent)
                del nevent
                return
        test = self.bugtags.findall(ievent.origtxt)
        if test:
            nevent = copy.copy(ievent)
            nevent.args = [test[0]]
            self.handle_show(bot, nevent)
            del nevent
            return

class Bugzilla(BugTracker):
    def __init__(self, url, autoshow):
        BugTracker.__init__(self, url, autoshow)
        self.bugurls = [
            re.compile(r'%s/show_bug.*id=(\d+)' % self.url)
            ]

    def list(self):
        csv = geturl2(self.list_url_summary())
        num = 0
        for line in csv.splitlines():
            try:
                num += int(line.split(',')[1])
            except ValueError:
                pass
        bugs = []
        if num > 100:
            bugs.append('showing 100-%d' % num)
        csv = geturl2(self.list_url())
        for line in csv.splitlines()[1:]:
            part = line.split(',')
            bugs.append('%s (%s)' % (part[0], part[1].replace('"', '')))
            if len(bugs) > 100:
                break
        return bugs        

    def list_url(self):
        return '%s/buglist.cgi?=&action=wrap&bug_file_loc=&bug_file_loc_type=allwordssubstr&bug_id=&bug_status=NEW&bug_status=ASSIGNED&bug_status=REOPENED&bugidtype=include&chfieldfrom=&chfieldto=Now&chfieldvalue=&email1=&email2=&emailtype1=substring&emailtype2=substring&field-1-0-0=bug_status&field0-0-0=noop&keywords=&keywords_type=allwords&long_desc=&long_desc_type=substring&query_format=advanced&remaction=&short_desc=&short_desc_type=allwordssubstr&status_whiteboard=&status_whiteboard_type=allwordssubstr&type-1-0-0=anyexact&type0-0-0=noop&value-1-0-0=NEW%%2CASSIGNED%%2CREOPENED&value0-0-0=&ctype=csv' % self.url

    def list_url_summary(self):
        return '%s/report.cgi?bug_file_loc=&bug_file_loc_type=allwordssubstr&bug_id=&bug_status=NEW&bug_status=ASSIGNED&bug_status=REOPENED&bugidtype=include&chfieldfrom=&chfieldto=Now&chfieldvalue=&email1=&email2=&emailtype1=substring&emailtype2=substring&field0-0-0=noop&keywords=&keywords_type=allwords&long_desc=&long_desc_type=substring&short_desc=&short_desc_type=allwordssubstr&status_whiteboard=&status_whiteboard_type=allwordssubstr&type0-0-0=noop&value0-0-0=&x_axis_field=&y_axis_field=bug_status&z_axis_field=&width=600&height=350&action=wrap&ctype=csv&format=table' % self.url

    def show(self, bugId):
        assert bugId.isdigit(), "bug id has to be a number"
        bugxml = geturl(self.show_url_xml(bugId))
        bugdom = xml.dom.minidom.parseString(bugxml)
        try:
            bugset = bugdom.getElementsByTagName('bug')[0]
        except:
            raise BugTrackerNotFound('could not find bug tag')
        bugget = {'creation_ts': 'created', 'product': 'product', 'component': 'component',
            'version': 'version', 'bug_status': 'status', 'resolution': 'resolution',
            'priority': 'priority', 'bug_severity': 'severity', 'reporter': 'reporter',
            'short_desc': 'description', 'assigned_to': 'assigned to'}
        data = {}
        for tag in bugget.keys():
            try:
                value = bugset.getElementsByTagName(tag)[0].firstChild.nodeValue
                data[bugget[tag]] = value
            except:
                pass
        return data

    def show_url(self, bugId):
        assert bugId.isdigit(), "bug id has to be a number"
        return '%s/show_bug.cgi?id=%s' % (self.url, bugId)

    def show_url_xml(self, bugId):
        assert bugId.isdigit(), "bug id has to be a number"
        return '%s/show_bug.cgi?ctype=xml&id=%s' % (self.url, bugId)

class Flyspray(BugTracker):
    def __init__(self, url, autoshow):
        BugTracker.__init__(self, url, autoshow)
        self.bugurls = [
            re.compile(r'%s/task/(\d+)' % self.url)
            ]

    def list(self):
        bugrss = geturl(self.list_url())
        bugdom = xml.dom.minidom.parseString(bugrss)
        bugall = bugdom.getElementsByTagName('entry')
        bugs = []
        if bugall:
            for entry in bugall:
                try:
                    bugid = entry.getElementsByTagName('id')[0].firstChild.nodeValue.split(':')[2]
                    bugs.append(bugid)
                except:
                    pass
        return bugs

    def list_url(self):
        return '%s/feed.php?feed_type=atom&project=' % self.url

    def show(self, bugId):
        assert bugId.isdigit(), "bug id has to be a number"
        html = geturl2(self.show_url(bugId))
        data = {}
        stat = ''
        for line in html.splitlines():
            line = line.strip()
            if not line:
                continue
            elif '<td headers="category">' in line:
                stat = 'category'
            elif '<td headers="status">' in line:
                stat = 'status'
            elif '<td headers="assignedto">' in line:
                stat = 'assigned to'
            elif '<td headers="os">' in line:
                data['os'] = striphtml(line).strip()
            elif '<td headers="severity">' in line:
                data['severity'] = striphtml(line).strip()
            elif '<td headers="priority">' in line:
                data['priority'] = striphtml(line).strip()
            elif '<td headers="reportedver">' in line:
                data['version'] = striphtml(line).strip()
            elif '<h2 class="summary' in line:
                stat = 'summary'
            elif '<a href="#comments">Comments (' in line:
                data['comments'] = line.split('(', 1)[1].split(')')[0]
            # stats
            elif stat:
                if stat in ['category', 'status', 'assigned to', 'summary']:
                    data[stat] = striphtml(line).strip()
                stat = ''
        return data

    def show_url(self, bugId):
        assert bugId.isdigit(), "bug id has to be a number"
        return '%s/task/%s' % (self.url, bugId)

class GoogleCode(BugTracker):
    def __init__(self, url, autoshow):
        BugTracker.__init__(self, url, autoshow)
        self.bugurls = [
            re.compile(r'%s/issues.*id=(\d+)' % self.url)
            ]

    def close(self, *args):
        raise BugTrackerNotSupported("google code needs authorization which I don't support")

    def list(self):
        html = geturl2(self.list_url())
        bugs = []
        for line in html.splitlines():
            if '<a href="detail?id' in line and '&amp;' in line:
                bug = line.split('id=')[1].split('&amp;')[0]
                if not bug in bugs:
                    bugs.append(bug)
        return bugs

    def list_url(self):
        return '%s/issues/list' % self.url

    def open(self, *args):
        raise BugTrackerNotSupported("google code needs authorization which I don't support")

    def show(self, bugId):
        assert bugId.isdigit(), "bug id has to ba a number"
        html = geturl2(self.show_url(bugId)).splitlines()
        data = {}
        stat = ''
        for line in html:
            line = line.strip()
            if not line:
                continue
            if line.startswith('<span class=h3 >'):
                data['summary'] = striphtml(line)
            elif line.startswith('<tr><th align=left>Status:'):
                stat = 'status'
            elif line.startswith('class=label><b>Type-</b>'):
                data['type'] = striphtml(line.split('</b>')[1])
            elif line.startswith('class=label><b>Priority-</b>'):
                data['priority'] = striphtml(line.split('</b>')[1])
            elif line.startswith('<span class=author>') and not data.has_key('author'):
                stat = 'author'
            elif line.startswith('<tr><th align=left>Owner:'):
                stat = 'owner'
            elif line.startswith('<span class="date" title="'):
                data['date'] = striphtml(line)
            elif striphtml(line) == '':
                pass
            # stats
            elif stat == 'author':
                data['reporter'] = striphtml(line)
                stat = ''
            elif stat == 'owner':
                data['owner'] = striphtml(line)
                stat = ''
            elif stat == 'status':
                data['status'] = striphtml(line)
                stat = ''
        return data

    def show_url(self, bugId):
        assert bugId.isdigit(), "bug id has to be a number"
        return '%s/issues/detail?id=%s' % (self.url.rstrip('/'), bugId)

class Mantis(BugTracker):
    def __init__(self, url, autoshow):
        BugTracker.__init__(self, url, autoshow)

    def show(self, bugId):
        assert bugId.isdigit(), "bug id has to be a number"
        html = geturl2(self.show_url(bugId))
        if 'APPLICATION ERROR #1100' in html:
            raise BugTrackerNotFound('issue not found')
        data = {'notes': 0}
        stat = ''
        skip = 0
        for line in html.splitlines():
            line = line.strip().replace('\t', '')
            if skip > 0:
                skip -= 1
                continue
            elif not line:
                continue
            elif '<!-- Category -->' in line:
                skip = 1
                stat = 'category'
            elif '<!-- Severity -->' in line:
                skip = 1
                stat = 'severity'
            elif '<!-- Reproducibility -->' in line:
                skip = 1
                stat = 'reproducibility'
            elif '<!-- Reporter -->' in line:
                skip = 3
                stat = 'reporter'
            elif '<!-- Priority -->' in line:
                skip = 1
                stat = 'priority'
            elif '<!-- Resolution -->' in line:
                skip = 1
                stat = 'resolution'
            elif '<!-- Status -->' in line:
                skip = 3
                stat = 'status'
            elif '<!-- Summary -->' in line:
                skip = 4
                stat = 'summary'
            elif '<td class="bugnote-public">' in line:
                data['notes'] += 1
            # stats
            elif stat:
                if stat in ['category', 'severity', 'reproducibility', 'reporter',
                    'priority', 'resolution', 'status', 'summary']:
                    data[stat] = striphtml(line)
                stat = ''
        return data

    def show_url(self, bugId):
        assert bugId.isdigit(), "bug id has to be a number"
        return '%s/view.php?id=%s' % (self.url.rstrip('/'), bugId)

class Trac(BugTracker):
    def __init__(self, url, autoshow):
        BugTracker.__init__(self, url, autoshow)
        self.bugurls = [
            re.compile(r'%s/ticket/(\d+)' % self.url)
            ]

    def comments(self, bugId):
        assert bugId.isdigit(), "bug id has to be a number"
        bugrss = geturl(self.comments_url(bugId))
        bugdom = xml.dom.minidom.parseString(bugrss)
        bugall = bugdom.getElementsByTagName('item')
        comments = []
        if bugall:
            for item in bugall:
                title = item.getElementsByTagName('title')[0].firstChild.nodeValue
                if 'comment added' in title:
                    try:
                        author = item.getElementsByTagName('dc:creator')[0].firstChild.nodeValue
                    except IndexError:
                        author = 'anonymous'
                    comment = item.getElementsByTagName('description')[0].firstChild.nodeValue
                    comment = striphtml(comment.replace('\n', ' ')).strip()
                    while '  ' in comment:
                        comment = comment.replace('  ', ' ')
                    comments.append('%s: %s' % (author, comment))
        return comments    

    def comments_url(self, bugId):
        assert bugId.isdigit(), "bug id has to be a number"
        return '%s/ticket/%s?format=rss' % (self.url, bugId)

    def list(self):
        tsv = geturl2(self.list_url()).splitlines()
        bugs = []
	keys = map(lambda x: x.strip(), tsv[0].split())
	for item in tsv[1:]:
		data = {}
		part = item.split('\t')
		# dirty hack to allow unescaped multilines in trac
		if part[0].isdigit() and part[1].isdigit() and len(part) == len(keys):
			data = dict(zip(keys, part))
			bugs.append('%s (%s)' % (data['summary'], data['ticket']))
        return bugs

    def list_url(self):
        return '%s/report/1?format=tab&USER=anonymous' % self.url

    def show(self, bugId):
        assert bugId.isdigit(), "bug id has to be a number"
        tsv = geturl2(self.show_url(bugId)+'?format=tab').splitlines()
	keys = map(lambda x: x.strip(), tsv[0].split())
	part = tsv[1].split('\t')
	data = dict(zip(keys, part))
        return data

    def show_url(self, bugId):
        assert bugId.isdigit(), "bug id has to be a number"
        return '%s/ticket/%s' % (self.url, bugId)

    def open(self, reporter, message, type='defect', priority='minor'):
        postdata = {
            'reporter':        reporter,
            'summary':         message,
            'type':            type,
            'description':     message,
            'priority':        priority,
            'milestone':       '',
            'component':       '',
            'version':         '',
            'keywords':        '',
            'owner':           '',
            'cc':              '',
            # hidden stuff
            'action':          'create',
            'status':          'new',
            }

    def open_url(self):
        return '%s/newticket' % self.url

    def close(self, bugId, reporter, message, action='fixed'):
        actions = ['fixed', 'invalid', 'wontfix', 'duplicate', 'worksforme']
        # input check
        assert bugId.isdigit(), "bug id has to be a number"
        assert action in actions, "action has to be one of: %s" % ', '.join(actions)
        showdata = self.show(bugId) 
        postdata = {
            'reporter':         reporter,
            'comment':          message,
            'action':           'resolve',
            'type':             'defect',
            'resolve_resolution': action,
            'summary':          showdata['summary'],
            'priority':         showdata['priority'],
            'milestone':        showdata['milestone'],
            'component':        showdata['component'],
            'version':          showdata['version'],
            'keywords':         showdata['keywords'],
            'cc':               '',
            }
        postdata = urllib.urlencode(postdata)
        req = urllib2.Request('%s/ticket/%s' % (self.url, bugId), data=postdata)
        req.add_header('User-agent', useragent())
        return urllib2.urlopen(req).read()

class BugManager(Pdod):

    types = {
        'bugzilla': Bugzilla,
        'flyspray': Flyspray,
        'googlecode': GoogleCode,
        'mantis': Mantis,
        'trac': Trac
        }
    instances = {
        }

    def __init__(self):
        Pdod.__init__(self, os.path.join(datadir + os.sep + 'plugs' + os.sep \
+ 'bugtracker', 'bugtracker'))
        self.load()

    def add(self, bot, ievent, type, url):
        assert type in self.types.keys(), "unknown bugtracker type, supported: %s" % ', '.join(self.types.keys())
        if hasattr(bot, 'name'):
            botname = bot.name
        else:
            botname = bot
        if hasattr(ievent, 'channel'):
            channel = ievent.channel
        else:
            channel = ievent
        if not self.data.has_key(botname):
            self.data[botname] = {}
        self.data[bot.name][channel] = [url, type, False]
        self.save()
        self.start(bot, ievent)

    def check(self, bot, ievent):
        if not self.instances.has_key(bot.name):
            return False
        if not self.instances[bot.name].has_key(ievent.channel):
            return False
        return True

    def load(self):
        for bot in self.data.keys():
            for channel in self.data[bot].keys():
                self.start(bot, channel)
                
    def start(self, bot, ievent):
        if hasattr(bot, 'name'):
            botname = bot.name
        else:
            botname = bot
        if hasattr(ievent, 'channel'):
            channel = ievent.channel
        else:
            channel = ievent
        if not self.instances.has_key(botname):
            self.instances[botname] = {}
        data = self.data[botname][channel]
        if not len(data) == 3:
            data.append(False)
            self.data[botname][channel] = data
            self.save()
        self.instances[botname][channel] = self.types[data[1]](data[0], data[2])

    # "internals"
    def handle_tracker_info(self, bot, ievent):
        if not self.check(bot, ievent):
            ievent.reply('no bugtracker active on %s' % ievent.channel)
        else:
            data = self.data[bot.name][ievent.channel]
            ievent.reply('bugtracker type: %s .. url: %s .. autoshow: %s' % (data[1], data[0], str(data[2])))
    
    def handle_tracker_set(self, bot, ievent):
        if len(ievent.args) != 2:
            ievent.missing('<type> <url>')
            return
        try:
            self.add(bot, ievent, ievent.args[0], ievent.args[1])
            ievent.reply('ok')
        except AssertionError, e:
            ievent.reply('error: %s' % e) 

    def handle_autoshow(self, bot, ievent):
        if not self.check(bot, ievent):
            ievent.reply('no bugtracker configured')
            return
        self.data[bot.name][ievent.channel][2] = not self.data[bot.name][ievent.channel][2]
        try:
            self.instances[bot.name][ievent.channel].autoshow = self.data[bot.name][ievent.channel][2]
        except:
            pass
        self.save()
        ievent.reply('autoshow is now %s' % str(self.data[bot.name][ievent.channel][2]))

    # proxies
    def handle_close(self, bot, ievent):
        if not self.check(bot, ievent):
            ievent.reply('no bugtracker configured')
            return
        try:
            self.instances[bot.name][ievent.channel].handle_close(bot, ievent)
        except Exception, e:
            ievent.reply('failed to get bug information: %s' % str(e))

    def handle_comments(self, bot, ievent):
        if not self.check(bot, ievent):
            ievent.reply('no bugtracker configured')
            return
        #try:
	self.instances[bot.name][ievent.channel].handle_comments(bot, ievent)
        #except Exception, e:
        #    ievent.reply('failed to get bug information: %s' % str(e))
    
    def handle_list(self, bot, ievent):
        if not self.check(bot, ievent):
            ievent.reply('no bugtracker configured')
            return
        self.instances[bot.name][ievent.channel].handle_list(bot, ievent)
        return
        try:
            self.instances[bot.name][ievent.channel].handle_list(bot, ievent)
        except Exception, e:
            ievent.reply('failed to get bug information: %s' % str(e))

    def handle_open(self, bot, ievent):
        if not self.check(bot, ievent):
            ievent.reply('no bugtracker configured')
            return
        try:
            self.instances[bot.name][ievent.channel].handle_open(bot, ievent)
        except Exception, e:
            ievent.reply('failed to get bug information: %s' % str(e))

    def handle_show(self, bot, ievent):
        if not self.data.has_key(bot.name):
            ievent.reply('no bugtracker configured')
            return False
        try:
            self.instances[bot.name][ievent.channel].handle_show(bot, ievent)
        except Exception, e:
            ievent.reply('failed to fetch bug information: %s' % str(e))

    def cb_privmsg(self, bot, ievent):
        if not self.instances.has_key(bot.name):
            return 0
        if not self.instances[bot.name].has_key(ievent.channel):
            return 0
        self.instances[bot.name][ievent.channel].scan_privmsg(bot, ievent)

bugtrackers = BugManager()

def init():
    callbacks.add('PRIVMSG', bugtrackers.cb_privmsg)

def upgrade():
    convertpickle(datadir + os.sep + 'old' + os.sep + 'bugtracker', \
datadir + os.sep + 'plugs' + os.sep + 'bugtracker' + os.sep + 'bugtracker')

cmnds.add('bug-autoshow', bugtrackers.handle_autoshow, ['OPER'])
examples.add('bug-autoshow', 'toggle automatic retrieval of bug information', 'bug-autoshow')
cmnds.add('bug-close', bugtrackers.handle_close, ['OPER', 'BUGS'])
examples.add('bug-close', 'close a bug', 'bug-close 42 fixed Typo on line 23')
cmnds.add('bug-comments', bugtrackers.handle_comments, ['USER'])
examples.add('bug-comments', 'show all comments on a bug', 'bug-comments 42')
cmnds.add('bug-list', bugtrackers.handle_list, ['USER', 'BUGS'])
examples.add('bug-list', 'list all open bugs', 'bug-list')
cmnds.add('bug-show', bugtrackers.handle_show, ['USER'])
examples.add('bug-show', 'show a bug', 'bug-show 42')
cmnds.add('bugtracker-info', bugtrackers.handle_tracker_info, ['BUGSOPER', 'OPER'])
cmnds.add('bugtracker-set', bugtrackers.handle_tracker_set, ['BUGSOPER', 'OPER'])
examples.add('bugtracker-set', 'set a bugtracker for the channel', 'bugtracker-set gozerbot http://dev.gozerbot.org')
aliases.data['bug'] = 'bug-show'
 



# gozerplugs/plugs/powernick.py
#
#

""" relay log messages to nicks joined on the partyline (DCC CHAT) """

from gozerbot.callbacks import callbacks
from gozerbot.persist.persiststate import PlugState
from gozerbot.commands import cmnds
from gozerbot.users import users
from gozerbot.examples import examples
from gozerbot.partyline import partyline
from gozerbot.plughelp import plughelp
import gozerbot.generic

plughelp.add('powernick', 'plugin used to monitor log messages')

state = PlugState()
state.define('nicks', {}) # username is key


def logcallback(level, description, txt):
    global state
    for nick in state['nicks'].values():
        partyline.say_nick(nick, "[%s] %s" % (description, txt)) 

def init():
    if not logcallback in gozerbot.utils.log.logcallbacks:
        gozerbot.utils.log.logcallbacks.append(logcallback)

def shutdown():
    if logcallback in gozerbot.utils.log.logcallbacks:
        gozerbot.utils.log.logcallbacks.remove(logcallback)

def handle_powernick(bot, ievent):
    global state
    nick = ievent.nick
    if not partyline.is_on(nick):
        ievent.reply('%s is not joined on the partylist .. make a dcc chat\
 connection to the bot' % nick)
        return
    username = users.getname(ievent.userhost)
    state['nicks'][username] = nick
    state.save()
    ievent.reply('powernick set to %s' % nick)

cmnds.add('powernick', handle_powernick, 'OPER')
examples.add('powernick', 'set powernick (gets /msg of bot logs)', 'powernick')

def handle_powernickdel(bot, ievent):
    global state
    username = users.getname(ievent.userhost)
    try:
        del state['nicks'][username]
        state.save()
        ievent.reply('powernick removed')
    except (ValueError, KeyError):
         ievent.reply('no powernick set')
 
cmnds.add('powernick-del', handle_powernickdel, 'OPER')
examples.add('powernick-del', 'remove powernick (gets', 'powernick-del')

def handle_powernicklist(bot, ievent):
    ievent.reply(state['nicks'].values())

cmnds.add('powernick-list', handle_powernicklist, 'OPER')
examples.add('powernick-list', 'show power nicks in use', 'powernick-list')



# plugs/quote.py
#
#

__copyright__ = 'this file is in the public domain'
__depend__ = ['karma', ]
__depending__ = ['grab', ]

from gozerbot.tests import tests
from gozerbot.persist.persist import Persist
from gozerbot.utils.nextid import nextid
from gozerbot.commands import cmnds
from gozerbot.examples import examples
from gozerbot.datadir import datadir
from gozerbot.generic import lockdec, rlog, handle_exception
from gozerbot.plughelp import plughelp
from gozerbot.aliases import aliases
from gozerbot.config import config
from gozerplugs.karma import karma
import random, re, time, thread, os, types

from gozerbot.database.alchemy import Base, create_session, trans, create_all
from datetime import datetime
from time import localtime   
from sqlalchemy import Column, String, Integer, Text, DateTime, ForeignKey, Sequence
import sqlalchemy as sa

class Quotes(Base):
    __tablename__ = 'quotes'
    __table_args__ = {'useexisting': True }
    indx = Column('indx', Integer, Sequence('quotes_indx_seq', optional=True), primary_key=True)
    quote = Column('quote', Text, nullable=False)
    userhost = Column('userhost', String(255), ForeignKey('userhosts.userhost'), nullable=False)
    createtime = Column('createtime', DateTime, nullable=False)
    nick = Column('nick', String(255), nullable=False)

    def __init__(self, quote, userhost, createtime, nick):
        self.quote = quote
        self.userhost = userhost
        self.createtime = createtime
        self.nick = nick

## UPGRADE PART

def upgrade():
    rlog(10, 'quote', 'upgrading')
    teller = 0
    oldfile = datadir + os.sep + 'old' + os.sep + 'quotes'
    dbdir = datadir + os.sep + 'db' + os.sep + 'quote.db'
    if os.path.exists(dbdir):
        try:
            from gozerbot.database.db import Db
            db = Db(dbtype='sqlite')
            db.connect('db/quote.db')
            result = db.execute(""" SELECT * FROM quotes """)
            if result:
                for i in result:
                    try:
                        quotes.add(i[4], i[2], i[1], i[3])
                    except:
                        quotes.add(i[3], i[2], i[1], i[4])
                    teller += 1
        except Exception, ex:
            handle_exception()
        return teller
    try:
        from gozerbot.utils.generic import dosed   
        from gozerbot.compat.persist import Persist
        from gozerbot.compat.todo import Todoitem
        oldpersist = Persist(oldfile) 
        if not oldpersist or not oldpersist.data:
            return
        for item in oldpersist.data:
            quotes.add(item.nick, item.userhost, item.txt, item.time)
            teller += 1
    except IOError, ex:
        if "No such file" in str(ex):
            rlog(10, 'quote', 'nothing to upgrade')
    except Exception, ex:
        rlog(10, 'quote', "can't upgrade .. reason: %s" % str(ex))
        handle_exception()
    else:
        rlog(10, 'quote', "upgraded %s items" % teller)

# END UPGRADE PART

plughelp.add('quote', 'manage quotes')

class QuotesDb(object):

    """ quotes db interface """

    def size(self):
        """ return number of items """
        s = create_session()
        count = s.query(sa.func.count(Quotes.indx)).first()[0]
        return count

    @trans
    def add(self, session, nick, userhost, quote, ttime=None):
        """ add a quote """
        if ttime:
            try:
                ttime = float(ttime)
            except TypeError:
                pass
            t = datetime.fromtimestamp(ttime)
        else: 
            t = datetime.now()
        quote = Quotes(quote, userhost, t, nick)
        session.add(quote)
        #session.commit()
        return quote.indx

    @trans
    def delete(self, session, quotenr):
        quote = session.query(Quotes).filter(Quotes.indx==quotenr).first()
        if quote:
             session.delete(quote)
             return 1

    def random(self):
        """ get random quote """
        s = create_session()
        result = s.query(Quotes.indx)
        indices = []
        if not result:
            return []
        for i in result:
            indices.append(i[0])
        if indices:
            idnr = random.choice(indices)
            return self.idquote(idnr)

    def idquote(self, quotenr):
        """ get quote by id """
        s = create_session()
        quote = s.query(Quotes).filter(Quotes.indx==quotenr).first()
        if quote:
            return quote

    def whoquote(self, quotenr):
        """ get who quoted the quote """
        s = create_session()
        quote = s.query(Quotes).filter(Quotes.indx==quotenr).first()
        if quote:
            return quote

    def last(self, nr=1):
        """ get last quote """ 
        s = create_session()
        result = s.query(Quotes).order_by(Quotes.indx.desc()).limit(nr)
        if result:
            return result

    def search(self, what):
        """ search quotes """
        s = create_session()
        result = s.query(Quotes).filter(Quotes.quote.like('%%%s%%' % what)).all()
        if result:
            return result

    def searchlast(self, what, nr):
        """ search quotes """
        s = create_session()
        result = s.query(Quotes).filter(Quotes.quote.like('%%%s%%' % what)).order_by(Quotes.indx.desc()).limit(nr)
        if result:
            return result

quotes = QuotesDb()
assert(quotes)

def init():
    create_all('quote')


def size():
    """ return number of quotes """
    return quotes.size()

def search(what, queue):
    """ search the quotes """
    rlog(10, 'quote', 'searched for %s' % what)
    result = quotes.search(what)
    if result:
        for i in result:
            queue.put_nowait("#%s %s" % (i.indx, i.quote))

def handle_quoteupgrade(bot, ievent):
    if quotes.size():
         if '-f' in ievent.optionset:
             pass
         else:   
             ievent.reply('there are already quote items in the main database .. not upgrading')
             ievent.reply('use the -f option to force an upgrade')
             return
    ievent.reply('upgrading quotes')
    nritems = upgrade()
    ievent.reply('%s items upgraded' % nritems)

cmnds.add('quote-upgrade', handle_quoteupgrade, 'OPER', options={'-f': ''})

def handle_quoteadd(bot, ievent):
    """ quote-add <txt> .. add a quote """
    if not ievent.rest:
        ievent.missing("<quote>")
        return
    idnr = quotes.add(ievent.nick, ievent.userhost, ievent.rest)
    ievent.reply('quote %s added' % idnr)

cmnds.add('quote-add', handle_quoteadd, ['USER', 'QUOTEADD'], allowqueue=False)
examples.add('quote-add', 'quote-add <txt> .. add quote', 'quote-add mekker')
aliases.data['aq'] = 'quote-add'
tests.add('quote-add mekker','quote (\d+) added')

def handle_quotewho(bot, ievent):
    """ quote-who <nr> .. show who added a quote """
    try:
        quotenr = int(ievent.args[0])
    except IndexError:
        ievent.missing("<nr>")
        return
    except ValueError:
        ievent.reply("argument must be an integer")
        return
    result = quotes.whoquote(quotenr)
    if not result:
        ievent.reply('no who quote data available')
        return
    if result.createtime:
        if type(result.createtime) == types.LongType:
            ievent.reply('quote #%s was made by %s on %s' % (quotenr, \
result.nick, datetime.fromtimestamp(result.createtime).ctime()))
        else:
            ievent.reply('quote #%s was made by %s on %s' % (quotenr, \
result.nick, result.createtime.ctime()))
    else:
        ievent.reply('quote #%s was made by %s' % (quotenr, result.nick))

cmnds.add('quote-who', handle_quotewho, ['USER', 'WEB', 'CLOUD', 'ANONQUOTE'])
examples.add('quote-who', 'quote-who <nr> .. show who quote <nr>', \
'quote-who 1')
aliases.data['wq'] = 'quote-who'
tests.add('quote-who 1', 'quote #(\d+) was made by (\S+)')

def handle_quotedel(bot, ievent):
    """ quote-del <nr> .. delete quote by id """
    try:
        quotenr = int(ievent.args[0])
    except IndexError:
        ievent.missing('<nr>')
        return
    except ValueError:
        ievent.reply('argument needs to be an integer')
        return
    if quotes.delete(quotenr):
        ievent.reply('quote %s deleted' % quotenr)
    else:
        ievent.reply("can't delete quote with nr %s" % quotenr)

cmnds.add('quote-del', handle_quotedel, ['QUOTEDEL', 'OPER', 'QUOTE'])
examples.add('quote-del', 'quote-del <nr> .. delete quote', 'quote-del 2')
aliases.data['dq'] = 'quote-del'
tests.add('quote-add mekker', 'quote (\d+)').add('quote-del %s', 'quote %s deleted')

def handle_quotelast(bot, ievent):
    """ quote-last .. show last quote """
    search = ""
    try:
        (nr, search) = ievent.args
        nr = int(nr)  
    except ValueError:
        try:
            nr = ievent.args[0]
            nr = int(nr)
        except (IndexError, ValueError):
            nr = 1
            try:
                search = ievent.args[0]
            except IndexError:
                search = ""
    if nr < 1 or nr > 4:
        ievent.reply('nr needs to be between 1 and 4')
        return
    search = re.sub('^d', '', search)
    if search:
        quotelist = quotes.searchlast(search, nr)
    else:
        quotelist = quotes.last(nr)
    if quotelist != None:
        for quote in quotelist:
            try:
                qkarma = karma.get('quote %s' % quote.indx)
                if qkarma:
                    ievent.reply('#%s (%s) %s' % (quote.indx, qkarma, quote.quote))
                else:
                    ievent.reply('#%s %s' % (quote.indx, quote.quote))
            except AttributeError:
                    ievent.reply('#%s %s' % (quote.indx, quote.quote))
    else:
        ievent.reply("can't fetch quote")

cmnds.add('quote-last', handle_quotelast, ['USER', 'WEB', 'ANONQUOTE', 'CLOUD'])
examples.add('quote-last', 'show last quote', 'quote-last')
aliases.data['lq'] = 'quote-last'
tests.add('quote-last', '#(\d+)')

def handle_quote2(bot, ievent):
    """ quote-2 .. show 2 random quotes """
    quote = quotes.random()
    if quote:
        qkarma = karma.get('quote %s' % quote.indx)
        if qkarma:
            ievent.reply('#%s (%s) %s' % (quote.indx, qkarma, quote.quote))
        else:
            ievent.reply('#%s %s' % (quote.indx, quote.quote))
    else:
        ievent.reply('no quotes yet')
        return
    quote = quotes.random()
    if quote:
        qkarma = karma.get('quote %s' % quote.indx)
        if qkarma:
            ievent.reply('#%s (%s) %s' % (quote.indx, qkarma, quote.quote))
        else:
            ievent.reply('#%s %s' % (quote.indx, quote.quote))

cmnds.add('quote-2', handle_quote2, ['USER', 'WEB', 'ANONQUOTE', 'CLOUD'])
examples.add('quote-2', 'quote-2 .. show 2 random quotes', 'quote-2')
aliases.data['2q'] = 'quote-2'
tests.add('quote-2', '#(\d+)')
aliases.data['3q'] = "quote && quote && quote"

def handle_quoteid(bot, ievent):
    """ quote-id <nr> .. show quote by id """
    try:
        quotenr = int(ievent.args[0])
    except IndexError:
        ievent.missing('<nr>')
        return
    except ValueError:
        ievent.reply('argument must be an integer')
        return
    quote = quotes.idquote(quotenr)
    if quote:
        qkarma = karma.get('quote %s' % quote.indx)
        if qkarma:
            ievent.reply('#%s (%s) %s' % (quote.indx, qkarma, quote.quote))
        else:
            ievent.reply('#%s %s' % (quote.indx, quote.quote))
    else:
        ievent.reply("can't fetch quote with id %s" % quotenr)

cmnds.add('quote-id', handle_quoteid, ['USER', 'WEB', 'ANONQUOTE', 'CLOUD'])
examples.add('quote-id', 'quote-id <nr> .. get quote <nr>', 'quote-id 2')
aliases.data['iq'] = 'quote-id'
tests.add('quote-id 1', '#(\d+)')

def handle_quote(bot, ievent):
    """ quote .. show random quote """
    quote = quotes.random()
    if quote:
        qkarma = karma.get('quote %s' % quote.indx)
        if qkarma:
            ievent.reply('#%s (%s) %s' % (quote.indx, qkarma, quote.quote))
        else:
            ievent.reply('#%s %s' % (quote.indx, quote.quote))
    else:
        ievent.reply('no quotes yet')

cmnds.add('quote', handle_quote, ['USER', 'WEB', 'ANONQUOTE', 'CLOUD'])
examples.add('quote', 'show random quote', 'quote')
aliases.data['q'] = 'quote'
tests.add('quote', '#(\d+)')

def handle_quotesearch(bot, ievent):
    """ quote-search <txt> .. search quotes """
    if not ievent.rest:
        ievent.missing('<item>')
        return
    else:
        what = ievent.rest
        nrtimes = 3
    result = quotes.search(what)
    if result:
        if ievent.queues:
            res = []
            for quote in result:
                res.append('#%s %s' % (quote.indx, quote.quote))
            ievent.reply(res)
            return            
        if nrtimes > len(result):
            nrtimes = len(result)
        randquotes = random.sample(result, nrtimes)
        for quote in randquotes:
            qkarma = karma.get('quote %s' % quote.indx)
            if qkarma:
                ievent.reply('#%s (%s) %s' % (quote.indx, qkarma, quote.quote))
            else:
                ievent.reply("#%s %s" % (quote.indx, quote.quote))
    else:
        ievent.reply('no quotes found with %s' % what)

cmnds.add('quote-search', handle_quotesearch, ['USER', 'WEB', \
'ANONQUOTE', 'CLOUD'])
examples.add('quote-search', 'quote-search <txt> .. search quotes for <txt>'\
, 'quote-search bla')
aliases.data['sq'] = 'quote-search'
tests.add('quote-add gozerbot is tof').add('quote-search gozerbot', 'tof')

def handle_quotescount(bot, ievent):
    """ quote-count .. show number of quotes """
    ievent.reply('quotes count is %s' % quotes.size())

cmnds.add('quote-count', handle_quotescount, ['USER', 'WEB', \
'ANONQUOTE', 'CLOUD'])
examples.add('quote-count', 'count nr of quotes', 'quote-count')
aliases.data['cq'] = 'quote-count'
tests.add('quote-count', 'quotes count is (\d+)')

def handle_quotegood(bot, ievent):
    """ show top ten positive karma """
    result = karma.quotegood(limit=10)
    if result:
        resultstr = ""
        for i in result:
            if i[1] > 0:
                resultstr += "%s: %s " % (i[0], i[1])
        ievent.reply('quote goodness: %s' % resultstr)
    else:
        ievent.reply('quote karma void')

cmnds.add('quote-good', handle_quotegood, ['USER', 'WEB', 'ANONQUOTE', 'CLOUD'])
examples.add('quote-good', 'show top 10 quote karma', 'quote-good')
tests.add('quote 1++').add('quote-good', 'quote goodness')

def handle_quotebad(bot, ievent):
    """ show top ten negative karma """
    result = karma.quotebad(limit=10)
    if result:
        resultstr = ""
        for i in result:
            if i[1] < 0:
                resultstr += "%s: %s " % (i[0], i[1])
        ievent.reply('quote badness: %s' % resultstr)
    else:
        ievent.reply('quote karma void')

cmnds.add('quote-bad', handle_quotebad, ['USER', 'WEB', 'ANONQUOTE', 'CLOUD'])
examples.add('quote-bad', 'show lowest 10 quote karma', 'quote-bad')
tests.add('quote 2--').add('quote-bad', 'quote badness')



# plugs/shoutcast.py
#
#

""" shoutcast watcher """

__copyright__ = 'this file is in the public domain'

from gozerbot.persist.persist import Persist
from gozerbot.commands import cmnds
from gozerbot.generic import geturl, striphtml, rlog, lockdec
from gozerbot.datadir import datadir
from gozerbot.fleet import fleet
from gozerbot.examples import examples
from gozerbot.threads.thr import start_new_thread
from gozerbot.plughelp import plughelp
from gozerbot.persist.persistconfig import PersistConfig
import os, time, thread

plughelp.add('shoutcast', 'query a shoutcast server or periodically watch \
them')

cfg = PersistConfig()
cfg.define('scwatch', 0)
cfg.define('nodes', [])

sclock = thread.allocate_lock()
locked = lockdec(sclock)

class SCwatcher(object):

    def __init__(self):
        self.stop = False
        self.songsplayed = []

    def run(self):
        self.starttime = int(time.time())
        res = ""
        while cfg.get('scwatch') and not self.stop:
            time.sleep(1)
            if self.stop:
                break
            godo = []    
            for botname, channel, name, node, polltime in cfg.get('nodes'):
                if not (int(time.time()) - self.starttime) % int(polltime):
                    godo.append((botname, channel, name, node))
                    continue            
            if godo:
                rlog(0, 'shoutcast', 'running scan: %s' % str(godo))
                start_new_thread(self.doscan, (godo, ))        

    @locked
    def doscan(self, scanlist):
        for botname, channel, name, node in scanlist:            
            try:
                result = geturl('http://%s/7.html' % node)
            except Exception, ex:
                rlog(10, 'shoutcast', "can't get %s shoutcast data: %s" % \
(node, str(ex)))
                continue
            try:
                res = result.split(',')[6]
            except IndexError:
                rlog(10, 'shoutcast', "can't match %s shoutcast data" % node)
                continue
            song = striphtml(res).strip().replace('\n', '')
            bot = fleet.byname(botname)
            if bot and channel in bot.state['joinedchannels']:
                got = False
                for ttime, played in self.songsplayed:
                    if played == song:
                        got = True
                if not got:
                    self.songsplayed.append((time.time(), song))
                    bot.say(channel, "now playing on %s: %s" % (name, song))
                else:
                    for ttime, played in self.songsplayed:
                        if time.time() - ttime > 1800:
                            self.songsplayed.remove((ttime, played))

scwatcher = SCwatcher()
 
def init():
    if cfg.get('scwatch'):
         start_new_thread(scwatcher.run, ())
    return 1

def shutdown():
    rlog(10, 'shoutcast', 'shutting down watcher')
    scwatcher.stop = True

def handle_sc(bot, ievent):
    try:
        server = ievent.args[0]
    except IndexError:
        ievent.missing('<server>')
        return
    try:
        result = geturl('http://%s/7.html' % server)
    except Exception, ex:
        ievent.reply("can't get shoutcast data: %s" % str(ex))
        return
    try:
        res = result.split(',')[6]
    except IndexError:
        ievent.reply("can't extract shoutcast data")
        return
    ievent.reply(striphtml(res).strip())

cmnds.add('sc', handle_sc, 'USER')
examples.add('sc', 'sc <host:port> .. ask server:port for currently running \
song', 'sc stream1.jungletrain.net:8000')

def handle_sclist(bot, ievent):
    ievent.reply("shoutcast nodes: %s" % cfg.get('nodes'))
    
cmnds.add('sc-list', handle_sclist, 'OPER')
examples.add('sc-list', 'show list of watched shoutcast servers', 'sc-list')

def handle_scadd(bot, ievent):
    try:
        name, node, polltime = ievent.args
    except ValueError:
        ievent.missing('<name> <host:port> <polltime>')
        return
    try:
        polltime = int(polltime)
    except ValueError:
        ievent.reply('polltime needs to be an integer')
        return
    if polltime < 60:
        ievent.reply('polltime in min. 60 seconds')
        return
    if not cfg.get('scwatch'):
        cfg.set('scwatch', 1)
        scwatcher.stop = False    
        start_new_thread(scwatcher.run, ())
    cfg.append('nodes', [bot.name, ievent.channel, name, node, polltime])
    cfg.save()
    ievent.reply('%s added' % node)
    
cmnds.add('sc-add', handle_scadd, 'OPER')
examples.add('sc-add', 'sc-add <server:port> .. add server to watcher', \
'sc-add jungletrain stream1.jungletrain.net 180')

def handle_scdel(bot, ievent):
    try:
        name = ievent.args[0].lower()
    except IndexError:
        ievent.missing('<name>')
        return
    got = 0
    nodes = cfg.get('nodes')
    for i in range(len(nodes)-1, -1, -1):
        if nodes[i][2].lower() == name:
            del nodes[i]
            got = 1
    if got:
        cfg.save()
        ievent.reply('%s deleted' % name)
    else:
        ievent.reply('%s is not in nodeslist' % name)

cmnds.add('sc-del', handle_scdel, 'OPER')
examples.add('sc-del', 'sc-del <name> .. remove node <name> from watcher', \
'sc-del jungletrain')

def handle_scsetpolltime(bot, ievent):
    try:
        name, sec = ievent.args
        sec = int(sec)
    except ValueError:
        ievent.missing('<name> <secondstosleep>')
        return
    if sec < 60:
        ievent.reply('minimun is 60 seconds')
        return
    name = name.lower()
    got = False
    for i in cfg.get('nodes'):
        if i[2].lower() == name:
            i[4] = sec
            got = True
    if got:
        cfg.save()
        ievent.reply('poll time set to %s' % sec)
    else:
        ievent.reply('%s is not in nodes list' % name)
        
cmnds.add('sc-setpolltime', handle_scsetpolltime, 'OPER')
examples.add('sc-setpolltime', 'set poll time of the shoutcast watcher', \
'sc-setpolltime jungletrain 120')

def handle_scpolltime(bot, ievent):
    try:
        name = ievent.args[0].lower()
    except IndexError:
        ievent.missing('<name>')
        return
    for i in cfg.get('nodes'):
        if i[2].lower() == name:
            ievent.reply("polltime of %s is %s" % (name, i[4]))
            return
            
cmnds.add('sc-polltime', handle_scpolltime, 'OPER')
examples.add('sc-polltime', 'get polltime of <name>', 'sc-polltime \
jungletrain')

def handle_scstartwatch(bot, ievent):
    if cfg.get('scwatch'):
        ievent.reply('shoutcast watcher is already running')
        return
    scwatcher.stop = False
    cfg.set('scwatch', 1)
    start_new_thread(scwatcher.run, ())
    ievent.reply('watcher started')
    
cmnds.add('sc-startwatch', handle_scstartwatch, 'OPER')
examples.add('sc-startwatch', 'start the shoutcast watcher', 'sc-startwatch')

def handle_scstopwatch(bot, ievent):
    scwatcher.stop = True
    cfg.set('scwatch', 0)
    ievent.reply('watcher stopped')
    
cmnds.add('sc-stopwatch', handle_scstopwatch, 'OPER')
examples.add('sc-stopwatch', 'stop the shoutcast watcher', 'sc-stopwatch')



# gozerplugs/cloud.py
# 
#

__copyright__ = 'this file is in the public domain'
__gendocfirst__ = ['cloud-enable', 'cloud-startserver', 'cloud-boot', \
'cloud-joinall']
__gendoclast__ = ['cloud-stopserver', 'cloud-disable']

from gozerbot.generic import handle_exception, rlog, getpostdata, waitforqueue
from gozerbot.commands import cmnds
from gozerbot.examples import examples
from gozerbot.plughelp import plughelp
from gozerbot.threads.threadloop import ThreadLoop
from gozerbot.rest.server import RestServer, RestServerAsync, RestRequestHandler
from gozerbot.rest.client import RestClientAsync
from gozerbot.persist.persistconfig import PersistConfig
from gozerbot.plugins import plugins
from gozerbot.users import users
from gozerbot.threads.thr import start_new_thread
from gozerbot.fleet import fleet
from gozerbot.irc.ircevent import Ircevent
from gozerbot.aliases import aliasset
from gozerbot.persist.persiststate import PersistState
from gozerbot.datadir import datadir
from gozerbot.persist.pdod import Pdod

from simplejson import dumps

import socket, re, asyncore, time, random, Queue, os, select

plughelp.add('cloud', 'cloud network')

## UPGRADE PART

def upgrade():
    pass

##33 END UPGRADE PART

cfg = PersistConfig()
cfg.define('enable', 0)
cfg.define('wait', 5)
cfg.define('host' , socket.gethostbyname(socket.getfqdn()))
cfg.define('name' , socket.getfqdn())
cfg.define('port' , 10101)
cfg.define('disable', [])
cfg.define('booturl', 'http://gozerbot.org:10101/')
cfg.define('servermode', 0)

waitre = re.compile(' wait (\d+)', re.I)
hp = "%s:%s" % (cfg.get('host'), cfg.get('port'))
url = "http://%s" % hp

# CLIENT PART

class Client(RestClientAsync):

    def auth(self, userhost, perm):
        kwargs = {}
        kwargs['userhost'] = userhost
        kwargs['perm'] = perm
        self.sendpost(kwargs)

class Node(object):

    def __init__(self, name, url):
        self.name = name
        self.url = url
        self.client = Client(name, url)
        self.seen = time.time()
        self.synced = time.time()
        self.regtime = time.time()
        self.filter = []
        self.catch = []
        self.lasterrortime = None
        self.lasterror = ""
        self.insocket = None

    def __str__(self):
        return "name=%s url=<%s> seen=%s" % (self.name, self.url, self.regtime)

    def doget(self, mount, cb, *args, **kwargs):
        self.client = Client(self.url + mount, self.name).addcb(cb)
        self.client.get()
        return self.client

    def dopost(self, mount, cb, *args, **kwargs):
        self.client = Client(self.url + mount, self.name).addcb(cb)
        self.client.post(**kwargs)
        return self.client

    def ping(self, cb):
        self.doget('/gozernet/ping', cb) 

class Cloud(ThreadLoop):

    def __init__(self):
        ThreadLoop.__init__(self, 'cloud')
        self.datadir = datadir + os.sep + 'plugs' + os.sep + 'cloud'
        self.nodes = {}
        self.state = Pdod(self.datadir + os.sep + 'state')
        self.startup = Pdod(self.datadir + os.sep + 'startup')
        if not self.state.has_key('ignore'):
            self.state['ignore'] = []
        if not self.state.has_key('names'):
            self.state['names'] = {}
        if not self.startup.has_key('start'):
            self.startup['start'] = {}
        self.enabled = False
        self.running = False

    def addifping(self, name, url):
        if not url.endswith('/'):
            url += '/'
        node = Node(name, url)
        def cb(client, result):
            if result.error:
                rlog(0, url, 'failed to receive pong .. not adding: %s' % \
result.error)
                return    
            if 'pong' in result.data:
                self.add(name, url)
            else:
                rlog(0, url, 'invalid ping data')
        client = Client(url + 'gozernet/ping').addcb(cb)
        client.get()

    def add(self, name, url, persist=True):
        if not url.endswith('/'):
            url += '/'
        node = Node(name, url)
        self.nodes[url] = node 
        self.state.set('names', name,  url)
        if persist:
            self.persist(name, url)
        rlog(0, url, 'added %s node <%s>' % (name, url))
        return self.nodes[url]

    def start(self, regname, regport, booturl=None):
        ThreadLoop.start(self)
        self.enabled = True
        for name, url in self.startup['start'].iteritems():
            self.add(name, url)
        def cb(client, result):
            if result.error:
                rlog(10, booturl, 'boot error: %s' % result.error)
            else:
                rlog(10, booturl, 'boot result: %s' % result.data)
        self.boot(regname, regport, booturl, cbin=cb)

    def handle(self, *args):
        try:
            asyncore.loop()
        except select.error, ex:
            if 'Bad file descriptor' in str(ex):
                pass
            else:
                raise

    def persist(self, name, url):
        try:
            if self.startup['start'][name] == url:
                return
        except KeyError:
            pass
        self.startup.set('start', name, url)
        self.startup.save()

    def get(self, url):
        return self.nodes[url]

    def byname(self, name):
        try:
            url = self.state['names'][name]
            if url:
                return self.get(url)
        except KeyError:
            return None

    def remove(self, id):
        target = self.get(id)
        if not target:
            return
        del self.state['names'][target.name]
        del self.nodes[id]

    def unpersist(self, url):
        try:
            del self.startup['start'][url]
            self.startup.save()
        except KeyError:
            pass
            
    def ignore(self, url):
        self.state['ignore'].append(url)

    def unignore(self, url):
        self.state.data['ignore'].remove(url)
        self.state.save()

    def doget(self, mount, *args, **kwargs):
        self.put('go')
        cp = dict(self.nodes)
        for url, node in cp.iteritems():
            node.doget(mount, *args, **kwargs)

    def dopost(self, mount, *args, **kwargs):
        self.put('go')
        cp = dict(self.nodes)
        for url, node in cp.iteritems():
            node.dopost(mount, *args, **kwargs)

    def getnodes(self):
        result = []
        cp = dict(self.nodes)
        for url, node in cp.iteritems():
            result.append((node.name, url))
        return result

    def getname(self, url):
        return self.get(url).name

    def addrecord(self, regname, url, cbin=None):
        rlog(0, regname, 'sending addrecord request')
        def cb(client, result):
            if result.error:
                rlog(10, url, 'addrecord error: %s' % result.error)
            else:
                rlog(10, url, 'record added')
        client = Client(url + '/gozernet/+addrecord').addcb(cb)
        if cbin:
            client.addcb(cbin)
        client.post(name=regname)
        return client

    def join(self, regname, regport, url, cbin=None):
        rlog(0, regname, 'joining on port %s' % regport)
        def cb(client, result):
            if result.error:
                rlog(0, url, 'join error: %s' % result.error)
            else:
                rlog(0, url, 'joined')
        client = Client(url + '/gozernet/+join').addcb(cb)
        if cbin:
            client.addcb(cbin)
        client.post(name=regname, port=regport)
        self.addrecord(regname, url, cbin)
        return client

    def joinall(self, regname, regport, cbin=None):
        cp = dict(self.nodes)
        for url, node in cp.iteritems():
            self.join(regname, regport, url, cbin)
 
    def boot(self, regname, regport, url=None, cbin=None):
        rlog(10, 'cloud', 'booting')
        if url:
            self.sync(url, cbin)
        else:
            self.sync('http://gozerbot.org:10101', cbin)

    def fullboot(self, cbin=None):
        teller = 0
        threads = []
        cp = dict(self.nodes)
        for node in cp.values():
            self.sync(node.url, False, cbin)
            teller += 1
        return teller

    def sync(self, url, throw=True, cbin=None):
        """ sync cache with node """
        rlog(10, url, "syncing")
        def cb(client, result):
            if result.error:
                rlog(10, url, "can't sync: %s" % result.error)
                return
            got = []
            for node in result.data:
                gnode = self.add(node[0], node[1])
                gnode.synced = time.time()
                got.append(node[0])
            rlog(10, url, "added %s" % str(got))
        client = Client('%s/gozernet/nodes/' % url).addcb(cb)
        if cbin:
            client.addcb(cbin)
        client.get()

    def size(self):
        return len(self.nodes)

    def list(self):
        res = []
        cp = dict(self.nodes)
        for node in cp.values():
            res.append(str(node))
        return res

    def names(self): 
        return self.state['names'].keys()

# SERVER PART

def nodes_GET(server, request):
    if not cloud.running:
        return dumps("cloud not initialized")
    else:
        return dumps(cloud.getnodes())

def ping_GET(server, request):
    return dumps('pong')

def addrecord_POST(server, request):
    try:
        input = getpostdata(request)
        name = input['name']
    except KeyError:
        rlog(5, host, 'addrecord .. no name provided')
        return dumps('no name provided')
    rlog(10, 'cloud', 'addrecord request for %s (%s)' % (name, request.ip)) 
    try:
        if not users.exist(name):
            users.add(name, ["cloud@%s" % request.ip, ], ['CLOUD', ])
            return dumps("%s added" % name)
    except Exception, ex:
        pass

def join_POST(server, request):
    if not cloud.running:
        return dumps('cloud not initialized')
    try:
        (host, port) = request.client_address
    except:
        return dumps("can't determine host/port")
    try:
        input = getpostdata(request)
        port = input['port']
        name = input['name']
    except KeyError:
        rlog(0, host, 'join: no port number or name provided')
        return dumps('no port number/name provided')
    rlog(0, 'cloud', 'join request for %s %s' % (name, port))
    #if cloud.byname(name):
    #    rlog(10, host, '%s name already taken' % name)
    #    return dumps('node name already taken')
    hp = "%s:%s" % (host, port)
    url = 'http://%s/' % hp
    cloud.addifping(name, url)
    cloud.addrecord(cfg.get('name'), url)
    return dumps('node added')

def auth_POST(server, request):
    try:
        input = getpostdata(request)
        userhost = input['userhost']
        perm = input['perm']
    except KeyError:
        rlog(0, request.host, 'auth: no port number or name provided')
        return dumps('no port userhost/permission provided')
    return dumps(users.allowed(userhost, perm))

def dispatch_POST(server, request):
    """ dispatch request into the cloud """
    try:
        (host, port) = request.client_address
    except:
        return ["can't determine host/port", ]
    try:
        input = getpostdata(request)
        cmnd = input['cmnd']
    except KeyError:
        return dumps(['need cmnd value', ])
    try:
        channel = input['channel']
    except KeyError:
        channel = "#cloud"
    if not channel:
        channel = '#cloud'
    bot = fleet.getfirstbot()
    ievent = Ircevent()
    ievent.txt = cmnd  
    ievent.nick = 'cloud'
    ievent.userhost = "cloud@%s" % host
    ievent.channel = channel
    q = Queue.Queue()
    ievent.queues.append(q)
    ievent.speed = 3
    ievent.bot = bot
    result = []
    if plugins.woulddispatch(bot, ievent):
        start_new_thread(plugins.trydispatch, (bot, ievent))
    else:
        return dumps(["can't dispatch %s" % cmnd, ])
    result = waitforqueue(q, 10)
    if not result:
        return dumps(["no result", ])
    res = []
    for item in result:
        res.append(str(item))
    return dumps(res)

# AGGREGATOR PART

class Aggregator(ThreadLoop):

    def __init__(self):
        ThreadLoop.__init__(self, 'aggregator')
        self.results = {}
        self.outre = re.compile('(/S+)\:s(/S)+s')

    def handle(self, id, client, cmnd, result):
        if not self.results.has_key(id):
            self.results[id] = []
        self.results[id].append((client, cmnd, result))

    def get(self, id):
        try:
            result = self.results[id]
        except KeyError:
            return
        del self.results[id]
        return result

    def output(self, id, ievent):
        res = self.get(id)
        if res:
            out = []
            for r in res:
                client, cmnd, result = r
                if result.data:
                    if result.data not in out:
                        out.append(result.data)
            ievent.reply(out, dot=True, fromm=client.name)

    def aggregate(self, id, ievent):
        res = self.get(id)
        if res:
            agg = {}
            for r in res:
                client, cmnd, result = r
                if result.data:  
                    for item in result.data:
                       try:
                           name, nr = item.split()
                           if not agg.has_key(name):
                               agg[name] = int(nr)  
                           else:
                               agg[name] += int(nr)
                       except ValueError:
                           pass
            ievent.reply('results => ', agg)

# INIT ? SHUTDOWN PART / GLOBAL VARS

aggregator = None
server = None
cloud = None

def startcloud():
    global cloud
    try:
        aggregator.start()
        cloud = Cloud()
        cloud.add(cfg.get('name'), url)
        cloud.start(cfg.get('name'), cfg.get('port'), cfg.get('booturl'))
        #time.sleep(2)
        #cloud.joinall(cfg.get('name'), cfg.get('port'))
        #time.sleep(2)
        #rlog(10, 'cloud', 'total of %s nodes' % cloud.size())
        return cloud
    except Exception, ex:
        handle_exception()

def stopcloud():
    cloud and cloud.stop()
    aggregator and aggregator.stop()

def startserver():
    global server
    try:
        server = RestServerAsync((cfg.get('host'), cfg.get('port')), RestRequestHandler)
        if server:
            server.start()
            rlog(10, 'cloud', 'running at %s:%s' % (cfg.get('host'), cfg.get('port')))
            server.addhandler('/gozernet/nodes/', 'GET', nodes_GET)
            server.addhandler('/gozernet/ping/', 'GET', ping_GET)
            server.addhandler('/gozernet/+auth/', 'POST', auth_POST)
            server.addhandler('/gozernet/+join/', 'POST', join_POST)
            server.addhandler('/gozernet/+dispatch/', 'POST', dispatch_POST)
            server.addhandler('/gozernet/+addrecord/', 'POST', addrecord_POST)
            for mount in cfg.get('disable'):
                server.disable(mount)
    except socket.error, ex:
        rlog(10, 'cloud', str(ex))
    except Exception, ex:
        handle_exception()

def stopserver():
    try:
        if not server:
            rlog(10, 'cloud', 'server is already stopped')
            return
        server.shutdown()
        #server.server_close()
        #time.sleep(3)
    except Exception, ex:
        handle_exception()
        pass

def init():
    """ init the cloud plugin """
    if not cfg.get('enable'):
        return 1
    global aggregator
    aggregator = Aggregator()
    if cfg.get('servermode'):
        startserver()
    time.sleep(4)
    startcloud()

def shutdown():
    """ shutdown the cloud plugin """
    try:
        stopcloud()
        stopserver()
    except Exception, ex:
        handle_exception()

def size():
    """ return number of cloud nodes """
    return cloud.size()

# CLOUD COMMANDS

def handle_clouddispatch(bot, ievent):
    """ dispatch <cmnd> on nodes """ 
    if not cloud:
        ievent.reply('cloud is not initialised')
        return
    if not cfg.get('enable'):    
        ievent.reply('cloud is not enabled')
        return
    if not ievent.rest:
        ievent.missing('[--node <nodename>] [-d] [-e] <command>')
        return
    starttime = time.time()
    try:
        name = ievent.options['--node']
    except KeyError: 
        name = None  
    try:
        wait = int(ievent.options['--w'])
    except KeyError: 
        wait = 3
    except ValueError:
        ievent.reply("%s is not an integer" % ievent.options['--w'])
        return
    cmnd = ievent.rest
    id = ievent.nick + str(random.random())
    def cb(client, result):
        if result.error:   
            if '-e' in ievent.optionset:
                ievent.reply("%s: %s => %s" % (client.name, cmnd, result.error))
            return
        if '-d' in ievent.optionset:
            ievent.reply("%s: %s => " % (client.name, cmnd), result.data, dot=True)
        else:
            aggregator.put(id, client, cmnd, result)
    if name:
        node = cloud.byname(name)
        if not node:
            ievent.reply('there is no node named %s' % name)
            return
        ievent.reply('dispatching "%s" onto %s node - wait (%s)' % (cmnd, name, \
wait))
        node.dopost('gozernet/+dispatch', cb, cmnd=cmnd, channel=ievent.channel)
    else:
        ievent.reply('dispatching "%s" onto %s nodes - wait (%s)' % (cmnd, cloud.size(), wait))
        cloud.dopost('gozernet/+dispatch', cb, cmnd=cmnd, channel=ievent.channel)
    if '-d' in ievent.optionset:
        return
    time.sleep(wait)
    aggregator.output(id, ievent)

cmnds.add('dispatch', handle_clouddispatch, ['USER', ], allowqueue=False, \
threaded=True, options={'--node': '', '-e': '', '--w': '4', '-d': ''})
examples.add('dispatch', 'dispatch [-d] [-e] [--w <seconds to wait>] \
[--node <nodename>] <cmnd> .. execute <cmnd> in the cloud', \
'1) dispatch version 2) dispatch -d version 3) dispatch -e version 4) \
dispatch --w 10 version')
aliasset('d', 'dispatch')
aliasset('dd', 'dispatch -d')

def handle_cloudping(bot, ievent):
    """ do a ping on a cloud node """
    if not cloud:
        ievent.reply('cloud is not initialised')
        return
    if not cfg.get('enable'):
        ievent.reply('cloud is not enabled .. see cloud-enable')
        return
    try:
        name = ievent.args[0]
    except IndexError:
        ievent.missing('<name>')
        return
    node = cloud.byname(name)
    if not node:
        ievent.reply('there is not node named %s' % name)
        return
    def cb(client, result):
        if result.error:
            ievent.reply('%s is not alive: %s' % (name, result.error))
            return
        if 'pong' in result.data:
            ievent.reply('%s is alive' % name)
    node.ping(cb)
    ievent.closequeue = False

cmnds.add('cloud-ping', handle_cloudping, 'OPER', threaded=True)
examples.add('cloud-ping', 'ping a cloud node', 'cloud-ping gozerbot.org')

def handle_cloudlist(bot, ievent):
    """ cloud-list .. list all nodes in cache """
    if not cloud:
        ievent.reply('cloud is not initialised')
        return
    if not cfg.get('enable'):
        ievent.reply('cloud is not enabled')
        return
    ievent.reply("cloud nodes: ", cloud.list(), dot=' \002||\002 ')

cmnds.add('cloud-list', handle_cloudlist, 'OPER')
examples.add('cloud-list', 'list nodes cache', 'cloud-list')

def handle_cloudenable(bot, ievent):
    """ cloud-enable .. enable the cloud """
    #if not cloud:
    #    ievent.reply('cloud is not initialised')
    #    return
    ievent.reply('enabling the cloud')
    cfg.set('enable', 1)
    cfg.save()
    plugins.reload('gozerplugs', 'cloud')
    ievent.reply('done')

cmnds.add('cloud-enable', handle_cloudenable, 'OPER', threaded=True)
examples.add('cloud-enable', 'enable the gozerbot cloud', 'cloud-enable')

def handle_clouddisable(bot, ievent):
    """cloud-disable .. disable the gozerbot cloud """
    if not cloud:
        ievent.reply('cloud is not initialised')
        return
    cfg.set('enable', 0)
    cfg.save()
    plugins.reload('gozerplugs', 'cloud')
    ievent.reply('cloud disabled')

cmnds.add('cloud-disable', handle_clouddisable, 'OPER', threaded=True)
examples.add('cloud-disable', 'disable the gozerbot cloud', 'cloud-disable')

def handle_cloudsync(bot, ievent):
    """ cloud-sync <node> .. sync nodes cache with node """ 
    if not cloud:
        ievent.reply('cloud is not initialised')
        return
    if not cfg.get('enable'):
        ievent.reply('cloud is not enabled')
        return
    try:
        url = ievent.args[0]
    except IndexError:
        ievent.missing('<url>')
        return
    start = time.time()
    def cb(client, result):
        teller = 0
        if cloud:
            for node in cloud.nodes.values():
                if node.synced > start:
                    teller += 1
        ievent.reply('%s nodes synced' % str(teller))
    cloud.sync(url, cbin=cb)
    
cmnds.add('cloud-sync', handle_cloudsync, 'OPER', threaded=True)
examples.add('cloud-sync', 'cloud-sync <url> .. sync with provided node', \
'cloud-sync http://gozerbot.org:10101')

def handle_cloudaddnode(bot, ievent):
    """ cloud-addnode <name> <url> .. add node to cache """
    global cloud
    if not cloud:
        ievent.reply('cloud is not initialised')
        return
    if not cfg.get('enable'):
        ievent.reply('cloud is not enabled')
        return
    try:
        (name, url) = ievent.args
    except ValueError:
        ievent.missing('<name> <url>')
        return
    client = Client(url)
    ip = socket.gethostbyname(client.host)
    url = "http://%s:%s" % (ip, client.port)
    cloud.add(name, url)
    cloud.persist(name, url)
    ievent.reply('%s added' % name)

cmnds.add('cloud-add', handle_cloudaddnode, 'OPER')
examples.add('cloud-add', 'cloud-add <name> <url> .. add a node to cache and \
persist it', 'cloud-add gozerbot.org http://gozerbot.org:10101')

def handle_cloudgetnode(bot, ievent):
    """ cloud-getnode .. show node of <name>  """
    if not cloud:
        ievent.reply('cloud is not initialised')
        return
    if not cfg.get('enable'):
        ievent.reply('cloud is not enabled')
        return
    try:
        name = ievent.args[0]
    except IndexError:
        ievent.missing('<name>')
        return
    node = cloud.byname(name)
    if not node:
        ievent.reply('there is no node named %s' % name)
        return
    ievent.reply(str(node))
 
cmnds.add('cloud-getnode', handle_cloudgetnode, 'OPER')
examples.add('cloud-getnode', 'cloud-getnode <name> .. get node of <name>', \
'cloud-getnode gozerbot.org')

def handle_cloudnames(bot, ievent):
    """ cloud-names .. show names with nodes in cache """
    if not cloud:
        ievent.reply('cloud is not initialised')
        return
    if not cfg.get('enable'):
        ievent.reply('cloud is not enabled')
        return
    ievent.reply("cloud node names: ", cloud.names(), dot=True)
 
cmnds.add('cloud-names', handle_cloudnames, 'OPER')
examples.add('cloud-names', 'show all node names', 'cloud-names')

def handle_cloudboot(bot, ievent):
    """ boot the cloud node cache """
    if not cloud:
        ievent.reply('cloud is not initialised')
        return
    if not cfg.get('enable'):
        ievent.reply('cloud is not enabled')
        return
    try:
        url = ievent.args[0]
    except IndexError:
        url = 'http://gozerbot.org:10101'
    start = time.time()
    teller = 0
    def cb(client, result):
        if result.error:
            ievent.reply('boot error: %s' % result.error)
            return 
        for node in cloud.nodes.values():
            if node.synced > start:
                teller += 1
        ievent.reply('booted %s nodes' % str(teller))
    cloud.boot(cfg.get('name'), cfg.get('port'), url, cbin=cb)
    ievent.closequeue=False
 
cmnds.add('cloud-boot', handle_cloudboot, 'OPER', threaded=True)
examples.add('cloud-boot', 'sync cloud nodes list with provided host', \
'1) cloud-boot 2) cloud-boot http://gozerbot.org:10101')

def handle_cloudfullboot(bot, ievent):
    """ cloud-fullboot .. boot from all nodes in cache """
    if not cloud:
        ievent.reply('cloud is not initialised')
        return
    if not cfg.get('enable'):
        ievent.reply('cloud is not enabled')
        return
    start = time.time()
    teller = 0
    def cb(client, result):
        if result.error:
            ievent.reply('boot error: %s' % result.error)
            return 
        for node in cloud.nodes.values():
            if node.synced > start:
                teller += 1
        ievent.reply('booted %s nodes' % str(teller))
    cloud.fullboot(cbin=cb)
    ievent.closequeue = False
 
cmnds.add('cloud-fullboot', handle_cloudfullboot, 'OPER')
examples.add('cloud-fullboot', 'do a boot on every node in the cloud node \
list', 'cloud-boot')

def handle_cloudremove(bot, ievent):
    if not cloud:
        ievent.reply('cloud is not initialised')
        return
    if not cfg.get('enable'):
        ievent.reply('cloud is not enabled')
        return
    if not ievent.rest:
        ievent.missing('<name>')
        return
    got = False
    try:
        url = cloud.state['names'][ievent.rest]
        if url:
            cloud.unpersist(url)
            cloud.remove(url)
            got = True
    except KeyError:
        ievent.reply('there is no %s cloud node' % ievent.rest)
        return
    except Exception, ex:
        ievent.reply('error removing %s: %s' % (ievent.rest, str(ex)))
        return
    if got:
        ievent.reply('%s node removed' % ievent.rest)
    else:
        ievent.reply('error removing %s node' % ievent.rest)

cmnds.add('cloud-remove', handle_cloudremove, 'OPER')
examples.add('cloud-remove', 'remove node with <name> from the cloud' , \
'cloud-remove gozerbot.org')

def handle_cloudjoin(bot, ievent):
    if not cloud:
        ievent.reply('cloud is not initialised')
        return
    if not cfg.get('enable'):
        ievent.reply('cloud is not enabled')
        return
    if not ievent.rest:
        ievent.missing('<name>')
        return
    def cb(client, result):
        if result.error:
            ievent.reply(result.error)
        else:
            ievent.reply(result.data)
    try:
        url = cloud.state['names'][ievent.rest]
        cloud.join(cfg.get('name'), cfg.get('port'), url, cb)
    except Exception, ex:
        ievent.reply('error joining %s: %s' % (ievent.rest, str(ex)))
        return
    ievent.reply('join reqeust sent to %s' % url)
    ievent.closequeue = False

cmnds.add('cloud-join', handle_cloudjoin, 'OPER', allowqueue=False)
examples.add('cloud-join', 'join node with <name>' , 'cloud-join gozerbot.org')

def handle_cloudjoinall(bot, ievent):
    if not cloud:
        ievent.reply('cloud is not initialised')
        return
    if not cfg.get('enable'):
        ievent.reply('cloud is not enabled')
        return
    def cb(client, result):
        if result.error:
            ievent.reply("%s: %s" %(client.name, result.error))
        else:
            ievent.reply("%s: %s" % (client.name, result.data))
    try:
        cloud.joinall(cfg.get('name'), cfg.get('port'), cb)
    except Exception, ex:
        handle_exception()
        ievent.reply('error joining %s: %s' % (ievent.rest, str(ex)))
        return
    ievent.reply('join requests sent')
    ievent.closequeue = False

cmnds.add('cloud-joinall', handle_cloudjoinall, 'OPER', allowqueue=False)
examples.add('cloud-joinall', 'join all nodes' , 'cloud-joinall')

def handle_cloudmeet(bot, ievent):
    if not cloud:
        ievent.reply('cloud is not initialised')
        return
    if not ievent.rest:
        ievent.missing('<nodename>')
        return
    name = ievent.rest
    node = cloud.byname(name)
    if not node:
       ievent.reply('%s is not a cloud node' % name)
       return
    if not node.client.host:
       ievent.reply("can't determine host of %s" % name)
       return
    try:
        if not users.exist(name):
            users.add(name, ["cloud@%s" % node.client.host, ], ['CLOUD', ])
            ievent.reply("%s (%s) added to database" % (name, node.client.host))
        else:
            ievent.reply("%s node already exists" % name)
    except Exception, ex:
        ievent.reply('error adding %s to the database: %s' % (name, str(ex)))

cmnds.add('cloud-meet', handle_cloudmeet, 'OPER')
examples.add('cloud-meet', 'cloud-meet <nodename>', 'cloud-meet gozerbot.org')

def handle_cloudallow(bot, ievent):
    """ cloud-allow .. allow a server mountpoint """
    if not cloud:
        ievent.reply('cloud is not initialised')
        return
    if not cfg.get('enable'):
        ievent.reply('cloud is not enabled')
        return
    if not ievent.rest:
        ievent.missing('<server mount>')
        return
    if server:
        server.enable(('gozernet', ievent.rest))
        ievent.reply('%s allowed' % ievent.rest)
    else:
        ievent.reply('cloud server not enabled')

cmnds.add('cloud-allow', handle_cloudallow, 'OPER')
examples.add('cloud-allow', 'allow execution of a server mountpoint', \
'cloud-allow auth')

def handle_clouddisallow(bot, ievent):
    """ cloud-disallow .. disallow a mount point"""
    if not cloud:
        ievent.reply('cloud is not initialised')
        return
    if not cfg.get('enable'):
        ievent.reply('cloud is not enabled')
        return
    if not ievent.rest:
        ievent.missing('<server mount>')
        return
    if server:
        server.disable(('gozernet', ievent.rest))
        ievent.reply('%s disallowed' % ievent.rest)
    else: 
        ievent.reply('cloud server not enabled')

cmnds.add('cloud-disallow', handle_clouddisallow, 'OPER')
examples.add('cloud-disallow', 'disallow execution of a server mountpoint', \
'cloud-disallow auth')

def handle_cloudstartserver(bot, ievent):
    """ cloud-startserver .. start the cloud server """
    if not cloud:
        ievent.reply('cloud is not initialised')
        return
    if not cfg.get('enable'):
        ievent.reply('cloud is not enabled')
        return
    cfg.set('servermode', 1)
    cfg.save()
    start_new_thread(startserver, ())
    ievent.reply('server thread started')
 
cmnds.add('cloud-startserver', handle_cloudstartserver, 'OPER')
examples.add('cloud-startserver', 'start the cloudserver', 'cloud-startserver')

def handle_cloudstopserver(bot, ievent):
    """ cloud-disallow .. disallow a mount point"""
    if not cloud:
        ievent.reply('cloud is not initialised')
        return
    cfg.set('servermode', 0)
    cfg.save()
    stopserver()
    ievent.reply('server disabled')
 
cmnds.add('cloud-stopserver', handle_cloudstopserver, 'OPER')
examples.add('cloud-stopserver', 'stop the cloud server', \
'cloud-stopserver')



# plugs/hex2ip.py
#
#

from gozerbot.commands import cmnds
from gozerbot.examples import examples
from gozerbot.plughelp import plughelp
import struct, socket

plughelp.add('hex2ip', 'convert a hexadecimal number to an ip addres')

def handle_hex2ip(bot, ievent):
    try:
        what = ievent.args[0]
    except IndexError:
        ievent.missing("<hexnr>")
        return
    try:
        ipint = int(what, 16)
        ipint = socket.ntohl(ipint)
        packed = struct.pack('l', ipint)
        ip = socket.inet_ntoa(str(packed))
    except Exception, ex:
        ievent.reply("can't make ipnr out of %s" % what)
        return
    ievent.reply('%s is %s' % (what, ip))
    try:
        hostname = socket.gethostbyaddr(ip)[0]
    except:
        ievent.reply("can't find hostname of %s" % ip)
        return
    ievent.reply('hostname is %s' % hostname)

cmnds.add('hex2ip', handle_hex2ip, 'USER')
examples.add('hex2ip', 'convert hexadecimal number to ip addres', 'hex2ip \
54008c8a')



# plugs/tel.py
#
#

""" telefone codes """

__copyright__ = 'this file is in the public domain'

from gozerbot.generic import geturl, striphtml, splittxt
from gozerbot.commands import cmnds
from gozerbot.aliases import aliases
from gozerbot.examples import examples
from gozerbot.plughelp import plughelp
import re, random

plughelp.add('tel', 'show country telephone codes')

codes = [
	("213", "ALGERIA", "DZ"),
	("102", "ANTIGUA BARBUDA", ""),
	("54", "ARGENTINA", "AR"),
	("374", "ARMENIA", "AM"),
	("297", "ARUBA", "AW"),
	("61", "AUSTRALIA", "AU"),
	("43", "AUSTRIA", "AT"),
	("994", "AZERBAIJAN", "AZ"),
	("103", "BAHAMAS", "BS"),
	("104", "BARBADOS", "BB"),
	("375", "BELARUS", "BY"),
	("32", "BELGIUM", "BE"),
	("105", "BERMUDA", "BM"),
	("591", "BOLIVIA", "BO"),
	("55", "BRAZIL", "BR"),
	("673", "BRUNEI", ""),
	("359", "BULGARIA", "BG"),
	("107", "CANADA", "CA"),
	("108", "CAYMAN ISLANDS", ""),
	("38", "CHILE", "CL"),
	("57", "COLOMBIA", "CO"),
	("506", "COSTA RICA", ""),
	("53", "CUBA", "CU"),
	("357", "CYPRUS", "CY"),
	("420", "CZECHREPUBLIC", ""),
	("45", "DENMARK", "DK"),
	("593", "ECUADOR", "EC"),
	("20", "EGYPT", "EG"),
	("503", "EL SALVADOR", ""),
	("358", "FINLAND", "FI"),
	("33", "FRANCE", "FR"),
	("49", "GERMANY", "DE"),
	("30", "GREECE", "GR"),
	("111", "GRENADA", "GD"),
	("502", "GUATEMALA", "GT"),
	("592", "GUYANA", "GY"),
	("504", "HONDURAS", "HN"),
	("852", "HONG KONG", ""),
	("36", "HUNGARY", "HU"),
	("91", "INDIA", "IN"),
	("62", "INDONESIA", "ID"),
	("353", "IRELAND", "IE"),
	("972", "ISRAEL", "IL"),
	("39", "ITALY", "IT"),
	("112", "JAMAICA", "JM"),
	("81", "JAPAN", "JP"),
	("962", "JORDAN", "JO"),
	("850", "KOREA", ""),
	("82", "KOREA SOUTH", ""),
	("965", "KUWAIT", "KW"),
	("352", "LUXEMBOURG", "LU"),
	("60", "MALAYSIA", "MY"),
	("52", "MEXICO", "MX"),
	("599", "NETH ANTILLES", ""),
	("31", "NETHERLANDS", "NL"),
	("64", "NEW ZEALAND", ""),
	("505", "NICARAGUA", "NI"),
	("47", "NORWAY", "NO"),
	("968", "OMAN", "OM"),
	("92", "PAKISTAN", "PK"),
	("507", "PANAMA", "PA"),
	("595", "PARAGUAY", "PY"),
	("51", "PERU", "PE"),
	("63", "PHILIPPINES", "PH"),
	("48", "POLAND", "PL"),
	("351", "PORTUGAL", "PT"),
	("852", "PRC", ""),
	("40", "ROMANIA", "RO"),
	("7", "RUSSIA", ""),
	("966", "SAUDI ARABIA", ""),
	("65", "SINGAPORE", "SG"),
	("421", "SLOVAKIA", "SK"),
	("386", "SLOVENIA", "SI"),
	("27", "SOUTH AFRICA", ""),
	("34", "SPAIN", "ES"),
	("115", "ST KITTS NEVIS", ""),
	("122", "ST LUCIA", ""),
	("116", "ST VINCENT", ""),
	("597", "SURINAME", "SR"),
	("46", "SWEDEN", "SE"),
	("41", "SWITZERLAND", "CH"),
	("866", "TAIWAN", ""),
	("66", "THAILAND", "TH"),
	("117", "TRINIDAD TOBAGO", ""),
	("90", "TURKEY", "TR"),
	("118", "TURKS CAICOS", ""),
	("44", "UNITED KINGDOM", "UK"),
	("380", "UKRAINE", "UA"),
	("971", "UNITED ARAB EMIRATES", ""),
	("598", "URUGUAY", "UY"),
	("1", "UNITED STATES OF AMERICA", "USA"),
	("58", "VENEZUELA", "VE"),
	("84", "VIETNAM", ""),
	("106", "VIRGIN IS BRITISH", ""),
	("123", "VIRGIN IS USA", ""),
	("967", "YEMAN", ""),
	("381", "YUGOSLAVIA", ""),
	("666", "NUMBER OF THE BEAST", ""),
]

def handle_tel(bot, ievent):
    if not ievent.rest:
        ievent.missing('<what>')
        return
    l = ievent.rest.upper()
    res = None
    for c in codes:
	if l in c:
	    res = "%s has dialing prefix %s" % (c[1], c[0])
	    break
    if not res:
	# match part of
	for c in codes:
	    if l in c[1]:
		res = "%s has dialing prefix %s" % (c[1], c[0])
		break
    if not res:
	res = "%s not found" % l
    ievent.reply(res)

cmnds.add('tel', handle_tel, 'USER')
examples.add('tel', 'show country telephone code', 'tel 31')



# plugs/rss.py
#
#

"""the rss mantra is of the following:

as OPER:

. add a url with rss-add
. start the watcher with rss-watch

now the user can start the bot sending messages of the feed to him with
rss-start. if an OPER gives a rss-start command in a channel data will be
send to the channel.
 """

__copyright__ = 'this file is in the public domain'
__gendocfirst__ = ['rss-add', 'rss-watch', 'rss-start']
__gendocskip__ = ['rss-dump', ]

from gozerbot.utils.generic import convertpickle
from gozerbot.persist.persist import PlugPersist
from gozerbot.generic import geturl2, handle_exception, rlog, lockdec, \
strippedtxt, fromenc, striphtml, useragent, jsonstring
from gozerbot.utils.rsslist import rsslist
from gozerbot.utils.statdict import Statdict
from gozerbot.utils.lazydict import LazyDict
from gozerbot.fleet import fleet
from gozerbot.commands import cmnds
from gozerbot.examples import examples
from gozerbot.datadir import datadir
from gozerbot.utils.dol import Dol
from gozerbot.persist.pdod import Pdod
from gozerbot.persist.pdol import Pdol
from gozerbot.plughelp import plughelp
from gozerbot.periodical import periodical
from gozerbot.aliases import aliasset
from gozerbot.users import users
from simplejson import loads
try:
    from gozerplugs.tinyurl import get_tinyurl
except:
    def get_tinyurl(url):
        return url

import gozerbot.compat.rss
import feedparser
import gozerbot.threads.thr as thr
import time, os, types, thread, socket, xml

plughelp.add('rss', 'manage rss feeds')

## UPGRADE PART

if not os.path.isdir(datadir + os.sep + 'plugs' + os.sep + 'rss'):
    os.mkdir(datadir + os.sep + 'plugs' + os.sep + 'rss')

def upgrade():
    rlog(10, 'rss', 'upgrading')
    teller = 0
    oldpickle = gozerbot.compat.rss.Rsswatcher(datadir + os.sep + \
'old' + os.sep + 'rss')
    for name, rssitem in oldpickle.data.iteritems():
        newrssitem = Rssitem(rssitem.name, rssitem.url, rssitem.itemslist, rssitem.watchchannels, \
rssitem.sleeptime, rssitem.running)
        watcher.data[name] = newrssitem
        teller += 1
    watcher.save()
    convertpickle(datadir + os.sep + 'rss.itemslist', datadir + os.sep + 'plugs' + \
os.sep + 'rss' + os.sep + 'rss.itemslist')
    convertpickle(datadir + os.sep + 'rss.markup', datadir + os.sep + 'plugs' + \
os.sep + 'rss' + os.sep + 'rss.markup')
    rlog(10, 'rss', 'upgraded %s items' % str(teller))

## END UPGRADE PART

savelist = []
possiblemarkup = {'separator': 'set this to desired item separator', \
'all-lines': "set this to 1 if you don't want items to be aggregated", \
'tinyurl': "set this to 1 when you want to use tinyurls", 'skipmerge': \
"set this to 1 if you want to skip merge commits", 'reverse-order': 'set \
this to 1 if you want the rss items displayed with oldest item first'}

def txtindicts(result, d):
    """ return lowlevel values in (nested) dicts """
    for j in d.values():
        if type(j) == types.DictType:
            txtindicts(result, j) 
        else:
            result.append(j)

def checkfordate(data, date):
    for item in data:
        try:
            d = item['updated']
        except KeyError:
            continue
        if date == d:
            return True
    return False

rsslock = thread.allocate_lock()
locked = lockdec(rsslock)

class RssException(Exception):
    pass

class Rss301(RssException):
    pass

class RssStatus(RssException):
    pass

class RssBozoException(RssException):
    pass

class RssNoSuchItem(RssException):
    pass

class Rssitem(LazyDict):

    """ item that contains rss data """

    def __init__(self, name="", url="", itemslist=[], watchchannels=[], \
sleeptime=30*60, running=0, d={}):
        if not d:
            LazyDict.__init__(self)
            self['name'] = name
            self['url'] = url
            self['itemslist'] = list(itemslist)
            self['watchchannels'] = list(watchchannels)
            self['sleeptime'] = int(sleeptime)
            self['running'] = int(running)
            self['stoprunning'] = 0
            self['botname'] = None
        else:
            LazyDict.__init__(self, d)

class Rssdict(PlugPersist):

    """ dict of rss entries """

    def __init__(self, filename):
        PlugPersist.__init__(self, filename)
        if not self.data:
            self.data = {}
        else:
            tmp = {}
            for name, item in self.data.iteritems():
                tmp[name] = Rssitem(d=item)
            self.data = tmp
        if self.data.has_key('itemslists'):
            del self.data['itemslists']
        self.itemslists = Pdol(datadir + os.sep + 'plugs' + os.sep + 'rss' + \
os.sep + filename + '.itemslists')
        self.handlers = {}
        self.results = {}
        self.jobids = {}
        self.rawresults = {}
        self.results = Dol()
        self.modified = {}
        self.etag = {}
        self.markup = Pdod(datadir + os.sep + 'plugs' + os.sep + 'rss' + \
os.sep + filename + '.markup')

    def size(self):
        """ return number of rss entries """
        return len(self.data)

    @locked
    def add(self, name, url):
        """ add rss item """
        rlog(10, 'rss', 'adding %s %s' % (name, url))
        self.data[name] = Rssitem(name, url, ['title', ])
        self.save()

    @locked
    def delete(self, name):
        """ delete rss item by name """
        target = None
        for j, i in self.data.iteritems():
           if i.name == name:
               target = i
        if target:
            try:
                target.running = 0
                del self.data[name]
                self.save()
            except:
                pass

    def byname(self, name):
        """ return rss item by name """
        try:
            return self.data[name]
        except:
            return

    def getdata(self, name):
        """ get data of rss feed """
        rssitem = self.byname(name)
        if rssitem == None:
            raise RssNoSuchItem("no %s rss item found" % name)
        try:
            modified = self.modified[name]
        except KeyError:
            modified = None
        try:
            etag = self.etag[name]
        except KeyError:
            etag = None
        rlog(9, 'rss', 'fetching %s' % rssitem.url)
        result = feedparser.parse(rssitem.url, modified=modified, etag=etag, \
agent=useragent())
        if result and result.has_key('bozo_exception'):
            rlog(1, 'rss', '%s bozo_exception: %s' % (name, \
result['bozo_exception']))
        try:
            status = result.status
        except AttributeError:
            status = 200
        if status != 200 and status != 301 and status != 302 and status != 304:
            raise RssStatus(status)
        try:
            self.modified[name] = result.modified
        except AttributeError:
            pass
        try:
            self.etag[name] = result.etag
        except AttributeError:
            pass
        if status == 304:
            rlog(0, 'rss', 'recieved 304 from %s rss feed' % name)
            return self.rawresults[name]
        else:
            self.rawresults[name] = result.entries
        return result.entries

class Rsswatcher(Rssdict):

    """ rss watchers """ 

    def __init__(self, filename):
        Rssdict.__init__(self, filename)
        
    def sync(self, name):
        result = self.getdata(name)
        if result:
            self.results[name] = result

    def changeinterval(self, name, interval):
        periodical.changeinterval(self.jobids[name], interval)

    @locked
    def startwatchers(self):
        """ start watcher threads """
        for j, z in self.data.iteritems():
            if z.running:
                thr.start_new_thread(self.watch, (z.name, ))

    @locked
    def stopwatchers(self):
        """ stop all watcher threads """
        for j, z in self.data.iteritems():
            if z.running:
                z.stoprunning = 1
        periodical.killgroup('rss')

    def dowatch(self, name, sleeptime=1800):
        rssitem = self.byname(name)
        if not rssitem:
            rlog(10, 'rss', "no %s rss item available" % name)
            return
        while 1:
            try:
                self.watch(name)
            except Exception, ex:
                rlog(100, 'rss', '%s feed error: %s' % (name, str(ex)))
                rlog(100, 'rss', '%s sleeping %s seconds for retry' % \
(name, sleeptime))
                time.sleep(sleeptime)
                if not rssitem.running:
                    break
            else:
                break

    def makeresult(self, name, target, data):
        res = []
        for j in data:
            tmp = {}
            if not self.itemslists[jsonstring([name, target])]:
                return []
            for i in self.itemslists[jsonstring([name, target])]:
                try:
                    tmp[i] = unicode(j[i])
                except KeyError:
                    continue
            res.append(tmp)
        return res

    def watch(self, name):
        """ start a watcher thread """
        # get basic data
        rlog(10, 'rss', 'trying %s rss feed watcher' % name)
        try:
            result = self.getdata(name)
        except RssException, ex:
            rlog(10, 'rss', "%s error: %s" % (name, str(ex)))
            result = []
        rssitem = self.byname(name)
        if not rssitem:
            raise RssNoItem()
        # poll every sleeptime seconds
        self.results[name] = result
        pid = periodical.addjob(rssitem.sleeptime, 0, self.peek, name, name)
        self.jobids[name] = pid
        rssitem.running = 1
        rssitem.stoprunning = 0
        self.data[name] = rssitem
        self.save()
        rlog(10, 'rss', 'started %s rss watch' % name)

    def makeresponse(self, name, res, channel, sep="\002||\002"):
        # loop over result to make a response
        result = u""
        itemslist = self.itemslists[jsonstring([name, channel])]
        if not itemslist:
            rssitem = self.byname(name)
            if not rssitem:
                return "no %s rss item" % name
            else:
                self.itemslists.extend(jsonstring([name, channel]), rssitem.itemslist)
                self.itemslists.save()
        for j in res:
            if self.markup.get(jsonstring([name, channel]), \
'skipmerge') and 'Merge branch' in j['title']:
                continue
            resultstr = u""
            for i in self.itemslists[jsonstring([name, channel])]:
                try:
                    item = unicode(j[i])
                    if not item:
                        continue
                    if item.startswith('http://'):
                        if self.markup.get(jsonstring([name, channel]), \
'tinyurl'):
                            try:
                                tinyurl = get_tinyurl(item)
                                if not tinyurl:
                                    resultstr += u"<%s> - " % item
                                else:
                                    resultstr += u"<%s> - " % tinyurl[0]
                            except Exception, ex:
                                handle_exception()
                                resultstr += u"<%s> - " % item
                        else:
                            resultstr += u"<%s> - " % item
                    else:
                        resultstr += u"%s - " % item.strip()
                except KeyError:
                    continue
            resultstr = resultstr[:-3]
            if resultstr:
                result += u"%s %s " % (resultstr, sep)
        return result[:-(len(sep)+2)]

    def peek(self, name, *args):
        rssitem = self.byname(name)
        if not rssitem or not rssitem.running or rssitem.stoprunning:
            return
        try:
            try:
                res = self.getdata(name)
            except socket.timeout:
                rlog(10, 'rss', 'socket timeout of %s' % name)
                return
            except RssException, ex:
                rlog(10, 'rss', '%s error: %s' % (name, str(ex)))
                return
            if not res:
                return
            res2 = []
            for j in res:
                try:
                    d = j['updated'] 
                except KeyError:
                    if j not in self.results[name]:
                        self.results[name].append(j)
                        res2.append(j)
                else:
                    if not checkfordate(self.results[name], d):
                        self.results[name].append(j)
                        res2.append(j)
            if not res2:
                return
            for item in rssitem.watchchannels:
                try:
                    (botname, channel) = item
                except:
                    try:
                        (botname, channel) = loads(item)
                    except:
                        rlog(10, 'rss', '%s is not in the format \
(botname,channel)' % str(item))
                bot = fleet.byname(botname)
                if not bot:
                    continue
                if self.markup.get(jsonstring([name, channel]), 'reverse-order'):
                    res2 = res2[::-1]
                if self.markup.get(jsonstring([name, channel]), 'all-lines'):
                    for i in res2:
                        response = self.makeresponse(name, [i, ], channel)
                        bot.say(channel, "\002%s\002: %s" % \
(rssitem.name, response), fromm=rssitem.name)
                else:
                    sep =  self.markup.get(jsonstring([name, channel]), 'separator')
                    if sep:
                        response = self.makeresponse(name, res2, channel, \
sep=sep)
                    else:
                        response = self.makeresponse(name, res2, channel)
                    bot.say(channel, "\002%s\002: %s" % (rssitem.name, \
response), fromm=rssitem.name)
        except Exception, ex:
            handle_exception(txt=name)

    @locked
    def stopwatch(self, name):
        """ stop watcher thread """
        for i, z in self.data.iteritems():
            if i == name:
                z.running = 0
                try:
                    del self.results[name]
                except KeyError:
                    pass
                self.data[i] = z
                self.save()
                try:
                    periodical.killjob(self.jobids[i])
                except KeyError:
                    pass
                return 1

    @locked
    def list(self):
        """ return of rss names """
        feeds = self.data.keys()
        return feeds

    @locked
    def runners(self):	
        """ show names/channels of running watchers """
        result = []
        for j, z in self.data.iteritems():
            if z.running == 1 and not z.stoprunning: 
                result.append((z.name, z.watchchannels))
        return result

    @locked
    def feeds(self, botname, channel):
        """ show names/channels of running watcher """
        result = []
        for j, z in self.data.iteritems():
            if jsonstring([botname, channel]) in z.watchchannels or [botname, channel] in z.watchchannels:
                result.append(z.name)
        return result

    @locked
    def url(self, name):
        """ return url of rssitem """
        for j, z in self.data.iteritems():
            if z.name == name:
                return z.url

    @locked
    def seturl(self, name, url):
        """ set url of rssitem """
        for j, z in self.data.iteritems():
            if z.name == name:
                z.url = url
                self.save()
                return True

    def scan(self, name):
        """ scan a rss url for used xml items """
        try:
            result = self.getdata(name)
        except RssException, ex:
            rlog(10, 'rss', '%s error: %s' % (name, str(ex)))
            return
        if not result:
            return
        keys = []
        for item in self.rawresults[name]:
             for key in item.keys():
                 keys.append(key)            
        statdict = Statdict()
        for key in keys:
            statdict.upitem(key)
        return statdict.top()  

    def search(self, name, item, search):
        res = []
        for result in self.rawresults[name]:
            try:
                title = result['title']
                txt = result[item]
            except KeyError:
                continue
            if search in title.lower():
                if txt:
                   res.append(txt)
        return res

    def searchall(self, item, search):
        res = []
        for name, results in self.rawresults.iteritems():
            for result in results:
                try:
                    title = result['title']
                    txt = result[item]
                except KeyError:
                    continue
                if search in title.lower():
                    if txt:
                        res.append("%s: %s" % (name, txt))
        return res

    def all(self, name, item):
        res = []
        for result in self.rawresults[name]:
            try:
                txt = result[item]
            except KeyError:
                continue
            if txt:
                res.append(txt)
        return res

watcher = Rsswatcher('rss')

def init(): 
    """ called after plugin import """
    thr.start_new_thread(watcher.startwatchers, ())
    return 1

def size():
    """ return number of watched rss entries """
    return watcher.size()

def shutdown():
    """ called before plugin import """
    watcher.stopwatchers()
    return 1
    
def handle_rssadd(bot, ievent):
    """ rss-add <name> <url> .. add a rss item """
    try:
        (name, url) = ievent.args
    except ValueError:
        ievent.missing('<name> <url>')
        return
    watcher.add(name, url)
    ievent.reply('rss item added')

cmnds.add('rss-add', handle_rssadd, 'OPER')
examples.add('rss-add', 'rss-add <name> <url> to the rsswatcher', 'rss-add \
gozerbot http://core.gozerbot.org/hg/dev/0.9/rss-log')

def handle_rssdel(bot, ievent):
    """ rss-del <name> .. delete a rss item """
    try:
        name = ievent.args[0]
    except IndexError:
        ievent.missing('<name>')
        return
    if watcher.byname(name):
        watcher.stopwatch(name)
        watcher.delete(name)
        ievent.reply('rss item deleted')
    else:
        ievent.reply('there is no %s rss item' % name)

cmnds.add('rss-del', handle_rssdel, 'OPER')
examples.add('rss-del', 'rss-del <name> .. remove <name> from the \
rsswatcher', 'rss-del mekker')

def handle_rsswatch(bot, ievent):
    """ rss-watch <name> .. start watcher thread """
    if not ievent.channel:
        ievent.reply('no channel provided')
    try:
        name, sleepsec = ievent.args
    except ValueError:
        try:
            name = ievent.args[0]
            sleepsec = 1800
        except IndexError:
            ievent.missing('<name> [secondstosleep]')
            return
    try:
        sleepsec = int(sleepsec)
    except ValueError:
        ievent.reply("time to sleep needs to be in seconds")
        return
    rssitem = watcher.byname(name)
    if rssitem == None:
        ievent.reply("we don't have a %s rss object" % name)
        return
    got = None
    if not rssitem.running or rssitem.stoprunning:
        if not rssitem.sleeptime:
            rssitem.sleeptime = sleepsec
        rssitem.running = 1
        rssitem.stoprunning = 0
        got = True
        watcher.save()
        try:
            watcher.watch(name)
        except Exception, ex:
            ievent.reply(str(ex))
            return
    if got:
        watcher.save()
        ievent.reply('watcher started')
    else:
        ievent.reply('already watching %s' % name)

cmnds.add('rss-watch', handle_rsswatch, 'OPER')
examples.add('rss-watch', 'rss-watch <name> [seconds to sleep] .. go \
watching <name>', '1) rss-watch gozerbot 2) rss-watch gozerbot 600')

def handle_rssstart(bot, ievent):
    """ rss-start <name> .. start a rss feed to a user """
    if not ievent.rest:
       ievent.missing('<feed name>')
       return
    name = ievent.rest
    rssitem = watcher.byname(name)
    if bot.jabber:
        if ievent.msg:
            target = ievent.userhost
        else:
            target = ievent.channel
    else:
        if users.allowed(ievent.userhost, ['OPER', ]) and not ievent.msg:
            target = ievent.channel
        else:
            target = ievent.nick
    if rssitem == None:
        ievent.reply("we don't have a %s rss object" % name)
        return
    if not rssitem.running:
        ievent.reply('%s watcher is not running' % name)
        return
    if jsonstring([bot.name, target]) in rssitem.watchchannels or [bot.name, target] in rssitem.watchchannels:
        ievent.reply('we are already monitoring %s on (%s,%s)' % \
(name, bot.name, target))
        return
    rssitem.watchchannels.append([bot.name, target])
    for item in rssitem.itemslist:
        watcher.itemslists.adduniq(jsonstring([name, target]), item)
    watcher.save()
    ievent.reply('%s started' % name)

cmnds.add('rss-start', handle_rssstart, ['RSS', 'USER'])
examples.add('rss-start', 'rss-start <name> .. start a rss feed \
(per user/channel) ', 'rss-start gozerbot')

def handle_rssstop(bot, ievent):
    """ rss-start <name> .. start a rss feed to a user """
    if not ievent.rest:
       ievent.missing('<feed name>')
       return
    name = ievent.rest
    rssitem = watcher.byname(name)
    if bot.jabber:
        target = ievent.userhost
    else:
        if users.allowed(ievent.userhost, ['OPER', ]) and not ievent.msg:
            target = ievent.channel
        else:
            target = ievent.nick
    if rssitem == None:
        ievent.reply("we don't have a %s rss feed" % name)
        return
    if not rssitem.running:
        ievent.reply('%s watcher is not running' % name)
        return
    try:
        rssitem.watchchannels.remove([bot.name, target])
    except ValueError:
        try:
            rssitem.watchchannels.remove([bot.name, target])
        except ValueError:
            ievent.reply('we are not monitoring %s on (%s,%s)' % \
(name, bot.name, target))
            return
    watcher.save()
    ievent.reply('%s stopped' % name)

cmnds.add('rss-stop', handle_rssstop, ['RSS', 'USER'])
examples.add('rss-stop', 'rss-stop <name> .. stop a rss feed \
(per user/channel) ', 'rss-stop gozerbot')

def handle_rsschannels(bot, ievent):
    """ rss-channels <name> .. show channels of rss feed """
    try:
        name = ievent.args[0]
    except IndexError:
        ievent.missing("<name>") 
        return
    rssitem = watcher.byname(name)
    if rssitem == None:
        ievent.reply("we don't have a %s rss object" % name)
        return
    if not rssitem.watchchannels:
        ievent.reply('%s is not in watch mode' % name)
        return
    result = []
    for i in rssitem.watchchannels:
        result.append(str(i))
    ievent.reply("channels of %s: " % name, result, dot=True)

cmnds.add('rss-channels', handle_rsschannels, 'OPER')
examples.add('rss-channels', 'rss-channels <name> .. show channels', \
'rss-channels gozerbot')

def handle_rssaddchannel(bot, ievent):
    """ rss-addchannel <name> [<botname>] <channel> .. add a channel to \
        rss item """
    try:
        (name, botname, channel) = ievent.args
    except ValueError:
        try:
            (name, channel) = ievent.args
            botname = bot.name
        except ValueError:
            try:
                name = ievent.args[0]
                botname = bot.name
                channel = ievent.channel
            except IndexError:
                ievent.missing('<name> [<botname>] <channel>')
                return
    rssitem = watcher.byname(name)
    if rssitem == None:
        ievent.reply("we don't have a %s rss object" % name)
        return
    if not rssitem.running:
        ievent.reply('%s watcher is not running' % name)
        return
    if jsonstring([botname, channel]) in rssitem.watchchannels or [botname, channel] in rssitem.watchchannels:
        ievent.reply('we are already monitoring %s on (%s,%s)' % \
(name, botname, channel))
        return
    rssitem.watchchannels.append([botname, channel])
    watcher.save()
    ievent.reply('%s added to %s rss item' % (channel, name))

cmnds.add('rss-addchannel', handle_rssaddchannel, 'OPER')
examples.add('rss-addchannel', 'rss-addchannel <name> [<botname>] <channel> \
..add <channel> or <botname> <channel> to watchchannels of <name>', \
'1) rss-addchannel gozerbot #dunkbots 2) rss-addchannel gozerbot main \
#dunkbots')

def handle_rsssetitems(bot, ievent):
    try:
        (name, items) = ievent.args[0], ievent.args[1:]
    except ValueError:
        ievent.missing('<name> <items>')
        return
    if bot.jabber or users.allowed(ievent.userhost, ['OPER', ]):
        target = ievent.channel.lower()
    else:
        target = ievent.nick.lower()
    if not watcher.byname(name):
        ievent.reply("we don't have a %s feed" % name)
        return
    watcher.itemslists.data[jsonstring([name, target])] = items
    watcher.itemslists.save()
    ievent.reply('%s added to (%s,%s) itemslist' % (items, name, target))

cmnds.add('rss-setitems', handle_rsssetitems, ['RSS', 'USER'])
examples.add('rss-setitems', 'set tokens of the itemslist (per user/channel)',\
 'rss-setitems gozerbot author author link pubDate')

def handle_rssadditem(bot, ievent):
    try:
        (name, item) = ievent.args
    except ValueError:
        ievent.missing('<name> <item>')
        return
    if bot.jabber or users.allowed(ievent.userhost, ['OPER', ]):
        target = ievent.channel.lower()
    else:
        target = ievent.nick.lower()
    if not watcher.byname(name):
        ievent.reply("we don't have a %s feed" % name)
        return
    watcher.itemslists.adduniq(jsonstring([name, target]), item)
    watcher.itemslists.save()
    ievent.reply('%s added to (%s,%s) itemslist' % (item, name, target))

cmnds.add('rss-additem', handle_rssadditem, ['RSS', 'USER'])
examples.add('rss-additem', 'add a token to the itemslist (per user/channel)',\
 'rss-additem gozerbot link')

def handle_rssdelitem(bot, ievent):
    try:
        (name, item) = ievent.args
    except ValueError:
        ievent.missing('<name> <item>')
        return
    if users.allowed(ievent.userhost, ['OPER', 'RSS']):
        target = ievent.channel.lower()
    else:
        target = ievent.nick.lower()
    if not watcher.byname(name):
        ievent.reply("we don't have a %s feed" % name)
        return
    try:
        watcher.itemslists.remove(jsonstring([name, target]), item)
        watcher.itemslists.save()
    except RssNoSuchItem:
        ievent.reply("we don't have a %s rss feed" % name)
        return
    ievent.reply('%s removed from (%s,%s) itemslist' % (item, name, target))

cmnds.add('rss-delitem', handle_rssdelitem, ['RSS', 'USER'])
examples.add('rss-delitem', 'remove a token from the itemslist \
(per user/channel)', 'rss-delitem gozerbot link')

def handle_rssmarkuplist(bot, ievent):
    ievent.reply('possible markups ==> ' , possiblemarkup, dot=True)

cmnds.add('rss-markuplist', handle_rssmarkuplist, 'USER')
examples.add('rss-markuplist', 'show possible markup entries', \
'rss-markuplist')

def handle_rssmarkup(bot, ievent):
    try:
        name = ievent.args[0]
    except IndexError:
        ievent.missing('<name>')
        return
    if users.allowed(ievent.userhost, ['OPER', ]):
        target = ievent.channel.lower()
    else:
        target = ievent.nick.lower()
    try:
        ievent.reply(str(watcher.markup[jsonstring([name, target])]))
    except KeyError:
        pass

cmnds.add('rss-markup', handle_rssmarkup, ['RSS', 'USER'])
examples.add('rss-markup', 'show markup list for a feed (per user/channel)', \
'rss-markup gozerbot')

def handle_rssaddmarkup(bot, ievent):
    try:
        (name, item, value) = ievent.args
    except ValueError:
        ievent.missing('<name> <item> <value>')
        return
    if users.allowed(ievent.userhost, ['OPER', ]):
        target = ievent.channel.lower()
    else:
        target = ievent.nick.lower()
    try:
        value = int(value)
    except ValueError:
        pass
    try:
        watcher.markup.set(jsonstring([name, target]), item, value)
        watcher.markup.save()
        ievent.reply('%s added to (%s,%s) markuplist' % (item, name, target))
    except KeyError:
        ievent.reply("no (%s,%s) feed available" % (name, target))

cmnds.add('rss-addmarkup', handle_rssaddmarkup, ['RSS', 'USER'])
examples.add('rss-addmarkup', 'add a markup option to the markuplist \
(per user/channel)', 'rss-addmarkup gozerbot all-lines 1')

def handle_rssdelmarkup(bot, ievent):
    try:
        (name, item) = ievent.args
    except ValueError:
        ievent.missing('<name> <item>')
        return
    if users.allowed(ievent.userhost, ['OPER', 'RSS']):
        target = ievent.channel.lower()
    else:
        target = ievent.nick.lower()
    try:
        del watcher.markup[jsonstring([name, target])][item]
    except (KeyError, TypeError):
        ievent.reply("can't remove %s from %s feed's markup" %  (item, name))
        return
    watcher.markup.save()
    ievent.reply('%s removed from (%s,%s) markuplist' % (item, name, target))

cmnds.add('rss-delmarkup', handle_rssdelmarkup, ['RSS', 'USER'])
examples.add('rss-delmarkup', 'remove a markup option from the markuplist \
(per user/channel)', 'rss-delmarkup gozerbot all-lines')

def handle_rssdelchannel(bot, ievent):
    """ rss-delchannel <name> [<botname>] <channel> .. delete channel \
        from rss item """
    botname = None
    try:
        (name, botname, channel) = ievent.args
    except ValueError:
        try:
            (name, channel) = ievent.args
            botname = bot.name
        except ValueError:
            try:
                name = ievent.args[0]
                botname = bot.name
                channel = ievent.channel
            except IndexError:
                ievent.missing('<name> [<botname>] [<channel>]')
                return
    rssitem = watcher.byname(name)
    if rssitem == None:
        ievent.reply("we don't have a %s rss object" % name)
        return
    if jsonstring([botname, channel]) in rssitem.watchchannels:
        rssitem.watchchannels.remove(jsonstring([botname, channel]))
        ievent.reply('%s removed from %s rss item' % (channel, name))
    elif [botname, channel] in rssitem.watchchannels:
        rssitem.watchchannels.remove([botname, channel])
        ievent.reply('%s removed from %s rss item' % (channel, name))
    else:
        ievent.reply('we are not monitoring %s on (%s,%s)' % (name, botname, \
channel))
        return
    watcher.save()

cmnds.add('rss-delchannel', handle_rssdelchannel, 'OPER')
examples.add('rss-delchannel', 'rss-delchannel <name> [<botname>] \
[<channel>] .. delete <channel> or <botname> <channel> from watchchannels of \
<name>', '1) rss-delchannel gozerbot #dunkbots 2) rss-delchannel gozerbot \
main #dunkbots')

def handle_rssstopwatch(bot, ievent):
    """ rss-stopwatch <name> .. stop a watcher thread """
    try:
        name = ievent.args[0]
    except IndexError:
        ievent.missing('<name>')
        return
    rssitem = watcher.byname(name)
    if not rssitem:
        ievent.reply("there is no %s rssitem" % name)
        return
    if not watcher.stopwatch(name):
        ievent.reply("can't stop %s watcher" % name)
        return
    ievent.reply('stopped %s rss watch' % name)

cmnds.add('rss-stopwatch', handle_rssstopwatch, 'OPER')
examples.add('rss-stopwatch', 'rss-stopwatch <name> .. stop polling <name>', \
'rss-stopwatch gozerbot')

def handle_rsssleeptime(bot, ievent):
    """ rss-sleeptime <name> .. get sleeptime of rss item """
    try:
        name = ievent.args[0]
    except IndexError:
        ievent.missing('<name>')
        return
    rssitem = watcher.byname(name)
    if rssitem == None:
        ievent.reply("we don't have a %s rss item" % name)
        return
    try:
        ievent.reply('sleeptime for %s is %s seconds' % (name, \
str(rssitem.sleeptime)))
    except AttributeError:
        ievent.reply("can't get sleeptime for %s" % name)

cmnds.add('rss-sleeptime', handle_rsssleeptime, 'OPER')
examples.add('rss-sleeptime', 'rss-sleeptime <name> .. get sleeping time \
for <name>', 'rss-sleeptime gozerbot')

def handle_rsssetsleeptime(bot, ievent):
    """ rss-setsleeptime <name> <seconds> .. set sleeptime of rss item """
    try:
        (name, sec) = ievent.args
        sec = int(sec)
    except ValueError:
        ievent.missing('<name> <seconds>')
        return
    if sec < 60:
        ievent.reply('min is 60 seconds')
        return
    rssitem = watcher.byname(name)
    if rssitem == None:
        ievent.reply("we don't have a %s rss item" % name)
        return
    rssitem.sleeptime = sec
    if rssitem.running:
        try:
            watcher.changeinterval(name, sec)
        except KeyError, ex:
            ievent.reply("failed to set interval: %s" % str(ex))
            return
    watcher.save()
    ievent.reply('sleeptime set')

cmnds.add('rss-setsleeptime', handle_rsssetsleeptime, 'OPER')
examples.add('rss-setsleeptime', 'rss-setsleeptime <name> <seconds> .. set \
sleeping time for <name> .. min 60 sec', 'rss-setsleeptime gozerbot 600')

def handle_rssget(bot, ievent):
    """ rss-get  <name> .. fetch rss data """
    try:
        name = ievent.args[0]
    except IndexError:
        ievent.missing('<name>')
        return
    channel = ievent.channel
    rssitem = watcher.byname(name)
    if rssitem == None:
        ievent.reply("we don't have a %s rss item" % name)
        return
    try:
        result = watcher.getdata(name)
    except Exception, ex:
        ievent.reply('%s error: %s' % (name, str(ex)))
        return
    if watcher.markup.get(jsonstring([name, channel]), 'reverse-order'):
        result = result[::-1]
    response = None
    go = watcher.markup.get(jsonstring([name, channel]), 'all-lines')
    if go:
        for i in result:
            response = watcher.makeresponse(name, [i, ], channel)
            bot.say(channel, "\002%s\002: %s" % (rssitem.name, response), \
fromm=rssitem.name)
    else:
        response = watcher.makeresponse(name, result, ievent.channel)
        if response:
            ievent.reply("results of %s: %s" % (name, response))
        else:
            ievent.reply("can't match watcher data")

cmnds.add('rss-get', handle_rssget, ['RSS', 'USER'], threaded=True)
examples.add('rss-get', 'rss-get <name> .. get data from <name>', \
'rss-get gozerbot')

def handle_rssrunning(bot, ievent):
    """ rss-running .. show which watchers are running """
    result = watcher.runners()
    resultlist = []
    teller = 1
    for i in result:
        resultlist.append("%s %s" % (i[0], i[1]))
    if resultlist:
        ievent.reply("running rss watchers: ", resultlist, nr=1)
    else:
        ievent.reply('nothing running yet')

cmnds.add('rss-running', handle_rssrunning, ['RSS', 'USER'])
examples.add('rss-running', 'rss-running .. get running rsswatchers', \
'rss-running')

def handle_rsslist(bot, ievent):
    """ rss-list .. return list of rss items """
    result = watcher.list()
    result.sort()
    if result:
        ievent.reply("rss items: ", result, dot=True)
    else:
        ievent.reply('no rss items yet')

cmnds.add('rss-list', handle_rsslist, ['RSS', 'USER'])
examples.add('rss-list', 'get list of rss items', 'rss-list')

def handle_rssurl(bot, ievent):
    """ rss-url <name> .. return url of rss item """
    try:
        name = ievent.args[0]
    except IndexError:
        ievent.missing('<name>')
        return
    result = watcher.url(name)
    if not result:
        ievent.reply("can't fetch url for %s" % name)
        return
    try:
        if ':' in result.split('@')[0]:
            if not ievent.msg:
                ievent.reply('run this command in a private message')
                return
            if not users.allowed(ievent.userhost, 'OPER'):
                ievent.reply('you need have OPER perms')
                return
    except (TypeError, ValueError):
        pass
    ievent.reply('url of %s: %s' % (name, result))

cmnds.add('rss-url', handle_rssurl, ['RSS', 'OPER'])
examples.add('rss-url', 'rss-url <name> .. get url from rssitem with \
<name>', 'rss-url gozerbot')

def handle_rssseturl(bot, ievent):
    """ rss-seturl <name> <url> .. set url of rss item """
    try:
        name = ievent.args[0]
        url = ievent.args[1]
    except IndexError:
        ievent.missing('<name> <url>')
        return
    oldurl = watcher.url(name)
    if not oldurl:
        ievent.reply("no %s rss item found" % name)
        return
    if watcher.seturl(name, url):
        watcher.sync(name)
        ievent.reply('url of %s changed from %s to %s' % (name, oldurl, \
url))
    else:
        ievent.reply('failed to set url of %s to %s' % (name, url))

cmnds.add('rss-seturl', handle_rssseturl, ['RSS', 'OPER'])
examples.add('rss-seturl', 'rss-seturl <name> <url> .. change url of rssitem',\
'rss-seturl gozerbot http://core.gozerbot.org/hg/dev/0.9/rss-log')

def handle_rssitemslist(bot, ievent):
    """ rss-itemslist <name> .. show itemslist of rss item """
    try:
        name = ievent.args[0]
    except IndexError:
        ievent.missing('<name>')
        return
    try:
        itemslist = watcher.itemslists[jsonstring([name, ievent.channel.lower()])]
    except KeyError:
        ievent.reply("no itemslist set for (%s, %s)" % (name, \
ievent.channel.lower()))
        return
    ievent.reply("itemslist of (%s, %s): " % (name, ievent.channel.lower()), \
itemslist, dot=True)

cmnds.add('rss-itemslist', handle_rssitemslist, ['RSS', 'USER'])
examples.add('rss-itemslist', 'rss-itemslist <name> .. get itemslist of \
<name> ', 'rss-itemslist gozerbot')

def handle_rssscan(bot, ievent):
    """ rss-scan <name> .. scan rss item for used xml items """
    try:
        name = ievent.args[0]
    except IndexError:
        ievent.missing('<name>')
        return
    if not watcher.byname(name):
        ievent.reply('no %s feeds available' % name)
        return
    try:
        result = watcher.scan(name)
    except Exception, ex:
        ievent.reply(str(ex))
        return
    if result == None:
        ievent.reply("can't get data for %s" % name)
        return
    res = []
    for i in result:
        res.append("%s=%s" % i)
    ievent.reply("tokens of %s: " % name, res)

cmnds.add('rss-scan', handle_rssscan, 'OPER')
examples.add('rss-scan', 'rss-scan <name> .. get possible items of <name> ', \
'rss-scan gozerbot')

def handle_rsssync(bot, ievent):
    """ rss-sync <name> .. sync rss item data """
    try:
        name = ievent.args[0]
    except IndexError:
        ievent.missing('<name>')
        return
    try:
        result = watcher.sync(name)
        ievent.reply('%s synced' % name)
    except Exception, ex:
        ievent.reply("%s error: %s" % (name, str(ex)))

cmnds.add('rss-sync', handle_rsssync, 'OPER')
examples.add('rss-sync', 'rss-sync <name> .. sync data of <name>', \
'rss-sync gozerbot')

def handle_rssfeeds(bot, ievent):
    """ rss-feeds <channel> .. show what feeds are running in a channel """
    try:
        channel = ievent.args[0]
    except IndexError:
        channel = ievent.printto
    try:
        result = watcher.feeds(bot.name, channel)
        if result:
            ievent.reply("feeds running in %s: " % channel, result, dot=True)
        else:
            ievent.reply('%s has no feeds running' % channel)
    except Exception, ex:
        ievent.reply("ERROR: %s" % str(ex))

cmnds.add('rss-feeds', handle_rssfeeds, ['USER', 'RSS'])
examples.add('rss-feeds', 'rss-feeds <name> .. show what feeds are running \
in a channel', '1) rss-feeds 2) rss-feeds #dunkbots')

def handle_rsslink(bot, ievent):
    try:
        feed, rest = ievent.rest.split(' ', 1)
    except ValueError:
        ievent.missing('<feed> <words to search>')
        return
    rest = rest.strip().lower()
    try:
        res = watcher.search(feed, 'link', rest)
        if not res:
            res = watcher.search(feed, 'feedburner:origLink', rest)
        if res:
            ievent.reply(res, dot=" \002||\002 ")
    except KeyError:
        ievent.reply('no %s feed data available' % feed)
        return

cmnds.add('rss-link', handle_rsslink, ['RSS', 'USER'])
examples.add('rss-link', 'give link of item which title matches search key', \
'rss-link gozerbot gozer')

def handle_rssdescription(bot, ievent):
    try:
        feed, rest = ievent.rest.split(' ', 1)
    except ValueError:
        ievent.missing('<feed> <words to search>')
        return
    rest = rest.strip().lower()
    res = ""
    try:
        ievent.reply(watcher.search(feed, 'description', rest), \
dot=" \002||\002 ")
    except KeyError:
        ievent.reply('no %s feed data available' % feed)
        return

cmnds.add('rss-description', handle_rssdescription, ['RSS', 'USER'])
examples.add('rss-description', 'give description of item which title \
matches search key', 'rss-description gozerbot gozer')

def handle_rssall(bot, ievent):
    try:
        feed = ievent.args[0]
    except IndexError:
        ievent.missing('<feed>')
        return
    try:
        ievent.reply(watcher.all(feed, 'title'), dot=" \002||\002 ")
    except KeyError:
        ievent.reply('no %s feed data available' % feed)
        return

cmnds.add('rss-all', handle_rssall, ['RSS', 'USER'])
examples.add('rss-all', "give titles of a feed", 'rss-all gozerbot')

def handle_rsssearch(bot, ievent):
    try:
        txt = ievent.args[0]
    except IndexError:
        ievent.missing('<txt>')
        return
    try:        
        ievent.reply(watcher.searchall('title', txt), dot=" \002||\002 ")
    except KeyError:
        ievent.reply('no %s feed data available' % feed)
        return

cmnds.add('rss-search', handle_rsssearch, ['RSS', 'USER'])
examples.add('rss-search', "search titles of all current feeds", \
'rss-search goz')

def handle_rssdump(bot, ievent):
    try:
        ievent.reply(str(watcher.rawresults[ievent.rest]))
    except Exception, ex:
        ievent.reply(str(ex))

cmnds.add('rss-dump', handle_rssdump, 'OPER')
examples.add('rss-dump', 'dump cached rss data', 'rss-dump')

def handle_rsspeek(bot, ievent):
    if not ievent.rest:
        ievent.missing('<feedname>')
        return
    watcher.peek(ievent.rest)

cmnds.add('rss-peek', handle_rsspeek, 'OPER')



# plugs/udp.py
#
#

"""
    the bot has the capability to listen for udp packets which it will use
    to /msg a given nick or channel.

    1) setup

        * do !reload udp to enable the udp plugin
        * call !udp-cfgsave to generate a config file in gozerdata/plugs/udp/config
        * edit this file .. esp. look at the udpallowednicks list
        * run ./bin/gozerbot-udp -s to generate clientside config file "udp-send"
        * edit this file
        * test with:

        ::

            echo "YOOO" | ./bin/gozerbot-udp

    2) limiter

        on IRC the bot's /msg to a user/channel are limited to 1 per 3 seconds so the
        bot will not excessflood on the server. you can use partyudp if you need no 
        delay between sent messages, this will use dcc chat to deliver the message.
        on jabber bots there is no delay

"""


__copyright__ = 'this file is in the public domain'

# IMPORT SECTION

from gozerbot.fleet import fleet
from gozerbot.generic import rlog, handle_exception, strippedtxt, lockdec
from gozerbot.config import config
from gozerbot.plughelp import plughelp
from gozerbot.partyline import partyline
from gozerbot.threads.thr import start_new_thread
from gozerbot.contrib.rijndael import rijndael
from gozerbot.persist.persistconfig import PersistConfig

import socket, re, time, Queue

# END IMPORT

plughelp.add('udp' , 'run the udp listen thread')

# VARS SECTION

cfg = PersistConfig()
cfg.define('udp', 1) # set to 0 to disnable
cfg.define('udpparty', 0)
cfg.define('udpipv6', 0)
cfg.define('udpmasks', ['192.168*', ])
cfg.define('udphost', "localhost")
cfg.define('udpport', 5500)
cfg.define('udpallow', ["127.0.0.1", ])
cfg.define('udpallowednicks', ["#gozerbot", "dunker"])
cfg.define('udppassword', "mekker", exposed=False)
cfg.define('udpseed', "blablablablablaz", exposed=False) # needs to be 16 chars wide
cfg.define('udpstrip', 1) # strip all chars < char(32)
cfg.define('udpsleep', 0) # sleep in sendloop .. can be used to delay pack
cfg.define('udpdblog', 0)
cfg.define('udpbot', 'default')

# END VARS

def _inmask(addr):

    """ check if addr matches a mask. """

    if not cfg['udpmasks']:
        return False
    for i in cfg['udpmasks']:
        i = i.replace('*', '.*')
        if re.match(i, addr):
            return True

class Udplistener(object):

    """ 
        listen for udp messages.

    """

    def __init__(self):
        self.outqueue = Queue.Queue()
        self.queue = Queue.Queue()
        self.stop = 0
        if cfg['udpipv6']:
            self.sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
        else:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        except:
            pass
        self.sock.setblocking(1)
        self.sock.settimeout(1)
        self.loggers = []

    def _outloop(self):

        """ 
            loop controling the rate of outputted messages.

            .. literalinclude:: ../../gozerplugs/udp.py
                :pyobject: Udplistener._outloop

        """

        rlog(5, 'udp', 'starting outloop')
        while not self.stop:
            (printto, txt) = self.outqueue.get()
            if self.stop:
                return
            self.dosay(printto, txt)
        rlog(5, 'udp', 'stopping outloop')

    def _handleloop(self):
        while not self.stop:
            (input, addr) = self.queue.get()
            if not input or not addr:
                continue
            if self.stop:
                break                
            self.handle(input, addr)
            if cfg['udpsleep']:
                time.sleep(cfg['udpsleep'] or 0.01)
        rlog(5, 'udp', 'shutting down udplistener')

    def _listen(self):
        """ listen for udp messages .. /msg via bot"""
        if not cfg['udp']:
            return
        try:
            fleet.startok.wait(5)
            bot = fleet.byname(cfg['udpbot'] or 'default')
            if not bot:
                rlog(10, 'udp', "can't find %s bot .. not starting" % cfg['udpbot'])
                return
            self.sock.bind((cfg['udphost'], cfg['udpport']))
            rlog(10, 'udp', 'udp listening on %s %s' % (cfg['udphost'], \
cfg['udpport']))
            self.stop = 0
        except IOError:
            handle_exception()
            self.sock = None
            self.stop = 1
            return
        # loop on listening udp socket
        bot.connectok.wait()
        while not self.stop:
            try:
                input, addr = self.sock.recvfrom(64000)
            except socket.timeout:
                continue
            except Exception, ex:
                try:
                    (errno, errstr) = ex
                except ValueError:
                    errno = 0
                    errstr = str(ex)
                if errno == 4:
                    rlog(10, self.name, str(ex))
                    break
                if errno == 35:
                    continue
                else:
                    handle_exception()
                    break
            if self.stop:
                break
            self.queue.put((input, addr))
        rlog(5, 'udp', 'shutting down main loop')

    def handle(self, input, addr):

        """ 
            handle an incoming udp packet. 

            :param input: txt in udp packet
            :type input: string
            :param addr: address info of udp packet
            :type add: (host, port) tuple

            .. literalinclude:: ../../gozerplugs/udp.py
                :pyobject: Udplistener.handle
        """

        if cfg['udpseed']:
            data = ""
            for i in range(len(input)/16):
                try:
                    data += crypt.decrypt(input[i*16:i*16+16])
                except Exception, ex:
                    rlog(10, 'udp', "can't decrypt: %s" % str(ex))
                    data = input
                    break
        else:
            data = input
        if cfg['udpstrip']:
            data = strippedtxt(data)
        # check if udp is enabled and source ip is in udpallow list
        if cfg['udp'] and (addr[0] in cfg['udpallow'] or \
_inmask(addr[0])):
            # get printto and passwd data
            header = re.search('(\S+) (\S+) (.*)', data)
            if header:
                # check password
                if header.group(1) == cfg['udppassword']:
                    printto = header.group(2)    # is the nick/channel
                    # check if printto is in allowednicks
                    if not printto in cfg['udpallowednicks']:
                        rlog(10, 'udp', "udp denied %s" % printto )
                        return
                    rlog(0, 'udp', str(addr[0]) +  " udp allowed")
                    text = header.group(3)    # is the text
                    self.say(printto, text)
                else:
                    rlog(10, 'udp', "can't match udppasswd from " + \
str(addr))
            else:
                rlog(10, 'udp', "can't match udp from " + str(addr[0]))
        else:
            rlog(10, 'udp', 'denied udp from ' + str(addr[0]))

    def say(self, printto, txt):

        """ 
            send txt to printto. 
 
            .. literalinclude:: ../../gozerplugs/udp.py
                :pyobject: Udplistener.say

        """

        self.outqueue.put((printto, txt))

    def dosay(self, printto, txt):

        """ 
            send txt to printto .. do some checks. 

           .. literalinclude:: ../../gozerplugs/udp.py
                :pyobject: Udplistener.dosay

        """

        if cfg['udpparty'] and partyline.is_on(printto):
            partyline.say_nick(printto, txt)
            return
        bot = fleet.byname(cfg['udpbot'])
        if not bot.jabber and not cfg['nolimiter']:
            time.sleep(3)
        bot.say(printto, txt)
        for i in self.loggers:
            i.log(printto, txt)

# the udplistener object
if cfg['udp']:
    udplistener = Udplistener()

# initialize crypt object if udpseed is set in config
if cfg['udp'] and cfg['udpseed']:
    crypt = rijndael(cfg['udpseed'])

def init():

    """ 
        init the udp plugin. 

        .. literalinclude:: ../../gozerplugs/udp.py
            :pyobject: init

    """

    if cfg['udp']:
        start_new_thread(udplistener._listen, ())
        start_new_thread(udplistener._handleloop, ())
        start_new_thread(udplistener._outloop, ())
    return 1
    
def shutdown():

    """ 
        shutdown the udp plugin.

        .. literalinclude:: ../../gozerplugs/udp.py
            :pyobject: init

    """

    if cfg['udp'] and udplistener:
        udplistener.stop = 1
        udplistener.outqueue.put_nowait((None, None))
        udplistener.queue.put_nowait((None, None))
    return 1

if cfg['udp'] and cfg['udpdblog']:

    from gozerbot.database.db import Db

    class Udpdblog:

        """ log udp data to database. """

        # see tables/udplog for table definition and add udpdblog = 1 to 
        # the config file
        db = Db()

        def log(self, printto, txt):

            """ do the actual logging. """

            try:
                res = self.db.execute("""INSERT into udplog(time,printto,txt)
values(%s,%s,%s) """, (time.time(), printto, txt))
            except Exception, ex:
                rlog(10, 'udp', 'failed to log to db: %s' % str(ex))
            return res

    udplistener.loggers.append(Udpdblog())
    rlog(10, 'udp', 'registered database udp logger')
